--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Driver Declarations
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	Command Handler Tables
--]]
EX_CMD = {}
PRX_CMD = {}
NOTIFY = {}
DEV_MSG = {}
LUA_ACTION = {}

--[[
Tables of functions
The following tables are function containers that are called within the following functions:	

	OnDriverInit()
		- first calls all functions contained within ON_DRIVER_EARLY_INIT table
		- then calls all functions contained within ON_DRIVER_INIT table
	OnDriverLateInit()
		- calls all functions contained within ON_DRIVER_LATEINIT table
	OnDriverUpdate()
		- calls all functions contained within ON_DRIVER_UPDATE table
	OnDriverDestroyed()
		- calls all functions contained within ON_DRIVER_DESTROYED table
	OnPropertyChanged()
		- calls all functions contained within ON_PROPERTY_CHANGED table
--]]

ON_DRIVER_INIT = {}
ON_DRIVER_EARLY_INIT = {}
ON_DRIVER_LATEINIT = {}
ON_DRIVER_UPDATE = {}
ON_DRIVER_DESTROYED = {}
ON_PROPERTY_CHANGED = {}

-- Constants
DEFAULT_PROXY_BINDINGID = 5001
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Common Driver Code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
--[[
	OnPropertyChanged
		Function called by Director when a property changes value.
	Parameters
		sProperty
			Name of property that has changed.
	Remarks
		The value of the property that has changed can be found with: Properties[sName]. Note
		that OnPropertyChanged is not called when the Property has been changed by the driver
		calling the UpdateProperty command, only when the Property is changed by the user from
		the Properties Page. This function is called by Director when a property changes value.
--]]
function OnPropertyChanged(sProperty)
	Dbg:Trace("OnPropertyChanged(" .. sProperty .. ") changed to: " .. Properties[sProperty])

	local propertyValue = Properties[sProperty]
	
	-- Remove any spaces (trim the property)
	local trimmedProperty = string.gsub(sProperty, " ", "")

	-- if function exists then execute (non-stripped)
	if (ON_PROPERTY_CHANGED[sProperty] ~= nil and type(ON_PROPERTY_CHANGED[sProperty]) == "function") then
		ON_PROPERTY_CHANGED[sProperty](propertyValue)
		return
	-- elseif trimmed function exists then execute
	elseif (ON_PROPERTY_CHANGED[trimmedProperty] ~= nil and type(ON_PROPERTY_CHANGED[trimmedProperty]) == "function") then
		ON_PROPERTY_CHANGED[trimmedProperty](propertyValue)
		return
	end
end

function ON_PROPERTY_CHANGED.DebugMode(propertyValue)
	gDebugTimer:KillTimer()
	Dbg:OutputPrint(propertyValue:find("Print") ~= nil)
	Dbg:OutputC4Log(propertyValue:find("Log") ~= nil)
	if (propertyValue == "Off") then return end
	gDebugTimer:StartTimer()
end

function ON_PROPERTY_CHANGED.DebugLevel(propertyValue)
	Dbg:SetLogLevel(tonumber(string.sub(propertyValue, 1, 1)))
end

---------------------------------------------------------------------
-- ExecuteCommand Code
---------------------------------------------------------------------
--[[
	ExecuteCommand
		Function called by Director when a command is received for this DriverWorks driver.
		This includes commands created in Composer programming.
	Parameters
		sCommand
			Command to be sent
		tParams
			Lua table of parameters for the sent command
--]]
function ExecuteCommand(sCommand, tParams)
	Dbg:Trace("ExecuteCommand(" .. sCommand .. ")")
	Dbg:Info(tParams)

	-- Remove any spaces (trim the command)
	local trimmedCommand = string.gsub(sCommand, " ", "")

	-- if function exists then execute (non-stripped)
	if (EX_CMD[sCommand] ~= nil and type(EX_CMD[sCommand]) == "function") then
		EX_CMD[sCommand](tParams)
	-- elseif trimmed function exists then execute
	elseif (EX_CMD[trimmedCommand] ~= nil and type(EX_CMD[trimmedCommand]) == "function") then
		EX_CMD[trimmedCommand](tParams)
	-- handle the command
	elseif (EX_CMD[sCommand] ~= nil) then
		QueueCommand(EX_CMD[sCommand])
	else
		Dbg:Alert("ExecuteCommand: Unhandled command = " .. sCommand)
	end
end

--[[
	Define any functions of commands (EX_CMD.<command>) received from ExecuteCommand that need to be handled by the driver.
--]]

function EX_CMD.Join ()	
	local url = "/BeoZone/Zone/Device/OneWayJoin"	
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MyButton1(tParams)
       local url = "/BeoZone/Zone/Snapshot/Activate/1"
       C4:urlPut(url, "", {["Content-Type"] = "application/json"})
end

function EX_CMD.MyButton2(tParams)
       local url = "/BeoZone/Zone/Snapshot/Activate/2"
       C4:urlPut(url, "", {["Content-Type"] = "application/json"})
end

function EX_CMD.MyButton3(tParams)
       local url = "/BeoZone/Zone/Snapshot/Activate/3"
       C4:urlPut(url, "", {["Content-Type"] = "application/json"})
end

function EX_CMD.MyButton4(tParams)
       local url = "/BeoZone/Zone/Snapshot/Activate/4"
       C4:urlPut(url, "", {["Content-Type"] = "application/json"})
end

function EX_CMD.StandPosition (tParams)
    local position = tParams.Position
	local url = "/BeoZone/Zone/Stand/Active"
	--local tData = {active = tStandCommandMap[position].digit}
	local tData = {active = tStandCommandMap[position].id}
	urlPut(url, tData, nil)
end

function EX_CMD.SpeakerGroup(tParams)
	local group = tParams.Group
	local url = "/BeoZone/Zone/Sound/SpeakerGroup/Active"
	--local tData = {active = tSpeakerGroupCommandMap[group].digit}
	local tData = {active = tSpeakerGroupCommandMap[group].id}
	urlPut(url, tData, nil)
end

function EX_CMD.PictureMode(tParams)
	local mode = tParams.Mode
	local url = "/BeoZone/Zone/Picture/Mode/Active"
	--local tData = {active = tPictureModeCommandMap[mode].digit}
	local tData = {active = tPictureModeCommandMap[mode].id}
	urlPut(url, tData, nil)
end

function EX_CMD.SoundMode(tParams)
	local mode = tParams.Mode
	local url = "/BeoZone/Zone/Sound/Mode/Active"
	--local tData = {active = tSoundModeCommandMap[mode].digit}
	local tData = {active = tSoundModeCommandMap[mode].id}
	urlPut(url, tData, nil)
end

function EX_CMD.PictureFormat2D3D(tParams)
	local pictureformat = tParams.Format
	local url = "/BeoZone/Zone/Picture/Format/2d3d/Active"
	local tData = {active = tPictureFormat2d3dCommandMap[pictureformat].id}
	urlPut(url, tData, nil)
end

function EX_CMD.Shuffle()
	local url = "/BeoZone/Zone/List/Shuffle"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuRoot ()
	local url = "/BeoZone/Zone/Menu/Root"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuContents ()
	local url = "/BeoZone/Zone/Menu/Contents"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuSetup ()
	local url = "/BeoZone/Zone/Menu/Setup"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuFavorites ()
	local url = "/BeoZone/Zone/Menu/Favorites"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuElectronicProgramGuide ()
	local url = "/BeoZone/Zone/Menu/ElectronicProgramGuide"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuVideoOnDemand ()
	local url = "/BeoZone/Zone/Menu/VideoOnDemand"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.MenuHbbTv ()
	local url = "/BeoZone/Zone/Menu/HbbTv"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.DeviceInformation ()
	local url = "/BeoZone/Zone/Device/Information"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.DeviceLanguages ()
	local url = "/BeoZone/Zone/Device/Languages"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.DeviceSubtiles()
	local url = "/BeoZone/Zone/Device/Subtiles"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function EX_CMD.Custom()
	local url = "/BeoZone/Zone/Device/Custom"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

--[[
	EX_CMD.LUA_ACTION
		Function called for any actions executed by the user from the Actions Tab in Composer.
--]]
function EX_CMD.LUA_ACTION(tParams)
	if tParams ~= nil then
		for cmd,cmdv in pairs(tParams) do
			if cmd == "ACTION" then
				if (LUA_ACTION[cmdv] ~= nil) then
					LUA_ACTION[cmdv]()
				else
					Dbg:Alert("Undefined Action")
					Dbg:Alert("Key: " .. cmd .. " Value: " .. cmdv)
				end
			else
				Dbg:Alert("Undefined Command")
				Dbg:Alert("Key: " .. cmd .. " Value: " .. cmdv)
			end
		end
	end
end

--[[
	LUA_ACTION.DisplayGlobals
		Implementation of Action "Display Globals". Executed when selecting the "Display Globals" action within Composer.
		Provided as an example for actions.
--]]
function LUA_ACTION.DisplayGlobals()
	print ("Global Variables")
	print ("----------------------------")

	for k,v in pairs(_G) do	-- globals`
		if not (type(v) == "function") then
			--print(k .. ":  " .. tostring(v))
			if (string.find(k, "^g%L")  == 1) then
				print(k .. ":  " .. tostring(v))
				if (type(v) == "table") then
					PrintTable(v, "   ")
				end
			end
		end
	end

	print ("")
end

function LUA_ACTION.GetMACAddress()
	local url = "/BeoDevice"
	local callback = update_mac_address
	urlGet(url, callback)
end

function PrintTable(tValue, sIndent)
	sIndent = sIndent or "   "
	for k,v in pairs(tValue) do
		print(sIndent .. tostring(k) .. ":  " .. tostring(v))
		if (type(v) == "table") then
			PrintTable(v, sIndent .. "   ")
		end
	end
end

---------------------------------------------------------------------
-- ReceivedFromProxy Code
---------------------------------------------------------------------
--[[
	ReceivedFromProxy(idBinding, sCommand, tParams)
		Function called by Director when a proxy bound to the specified binding sends a
		BindMessage to the DriverWorks driver.

	Parameters
		idBinding
			Binding ID of the proxy that sent a BindMessage to the DriverWorks driver.
		sCommand
			Command that was sent
		tParams
			Lua table of received command parameters
--]]
function ReceivedFromProxy(idBinding, sCommand, tParams)
	if (sCommand ~= nil) then
		if(tParams == nil)		-- initial table variable if nil
			then tParams = {}
		end
		Dbg:Trace("ReceivedFromProxy(): " .. sCommand .. " on binding " .. idBinding .. "; Call Function " .. sCommand .. "()")
		Dbg:Info(tParams)

		if (PRX_CMD[sCommand]) ~= nil then
			PRX_CMD[sCommand](idBinding, tParams)
		else
			Dbg:Alert("ReceivedFromProxy: Unhandled command = " .. sCommand)
		end
	end
end

---------------------------------------------------------------------
-- Notification Code
---------------------------------------------------------------------
-- notify with parameters
function SendNotify(notifyText, Parms, bindingID)
	C4:SendToProxy(bindingID, notifyText, Parms, "NOTIFY")
end

-- A notify with no parameters
function SendSimpleNotify(notifyText, ...)
	bindingID = select(1, ...) or DEFAULT_PROXY_BINDINGID
	C4:SendToProxy(bindingID, notifyText, {}, "NOTIFY")
end

---------------------------------------------------------------------
-- Initialization/Destructor Code
---------------------------------------------------------------------
--[[
	OnDriverInit
		Invoked by director when a driver is loaded. This API is provided for the driver developer to contain all of the driver
		objects that will require initialization.
--]]
function OnDriverInit()
	C4:ErrorLog("INIT_CODE: OnDriverInit()")
	-- Call all ON_DRIVER_EARLY_INIT functions.
	for k,v in pairs(ON_DRIVER_EARLY_INIT) do
		if (ON_DRIVER_EARLY_INIT[k] ~= nil and type(ON_DRIVER_EARLY_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_EARLY_INIT." .. k .. "()")
			ON_DRIVER_EARLY_INIT[k]()
		end
	end

	-- Call all ON_DRIVER_INIT functions
	for k,v in pairs(ON_DRIVER_INIT) do
		if (ON_DRIVER_INIT[k] ~= nil and type(ON_DRIVER_INIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_INIT." .. k .. "()")
			ON_DRIVER_INIT[k]()
		end
	end

	-- Fire OnPropertyChanged to set the initial Headers and other Property global sets, they'll change if Property is changed.
	for k,v in pairs(Properties) do
		OnPropertyChanged(k)
	end
end

--[[
	OnDriverUpdate
		Invoked by director when an update to a driver is requested. This request can occur either by adding a new version of a driver
		through the driver search list or right clicking on the driver and selecting "Update Driver" from within ComposerPro.
		Its purpose is to initialize all components of the driver that are reset during a driver update.
--]]
function OnDriverUpdate()
	C4:ErrorLog("INIT_CODE: OnDriverUpdate()")
	
	-- Call all ON_DRIVER_UPDATE functions
	for k,v in pairs(ON_DRIVER_UPDATE) do
		if (ON_DRIVER_UPDATE[k] ~= nil and type(ON_DRIVER_UPDATE[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_UPDATE." .. k .. "()")
			ON_DRIVER_UPDATE[k]()
		end
	end
end

--[[
	OnDriverLateInit
		Invoked by director after all drivers in the project have been loaded. This API is provided
		for the driver developer to contain all of the driver objects that will require initialization
		after all drivers in the project have been loaded.
--]]
function OnDriverLateInit()
	C4:ErrorLog("INIT_CODE: OnDriverLateInit()")
	
	-- Call all ON_DRIVER_LATEINIT functions
	for k,v in pairs(ON_DRIVER_LATEINIT) do
		if (ON_DRIVER_LATEINIT[k] ~= nil and type(ON_DRIVER_LATEINIT[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_LATEINIT." .. k .. "()")
			ON_DRIVER_LATEINIT[k]()
		end
	end
end


--[[
	OnDriverDestroyed
		Function called by Director when a driver is removed. Release things this driver has allocated such as timers.
--]]
function OnDriverDestroyed()
	C4:ErrorLog("INIT_CODE: OnDriverDestroyed()")
	-- Call all ON_DRIVER_DESTROYED functions
	for k,v in pairs(ON_DRIVER_DESTROYED) do
		if (ON_DRIVER_DESTROYED[k] ~= nil and type(ON_DRIVER_DESTROYED[k]) == "function") then
			C4:ErrorLog("INIT_CODE: ON_DRIVER_DESTROYED." .. k .. "()")
			ON_DRIVER_DESTROYED[k]()
		end
	end
end
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Device Communication Code (Serial, Network)
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Constants
SERIAL_BINDING_ID = 1
IR_BINDING_ID = 2
NETWORK_BINDING_ID = 6001
NETWORK_PORT = 8080
COMMAND_QUEUE_SIZE = 100

--[[
	This will be called by the base template code from its OnDriverInit function.
--]]
function ON_DRIVER_INIT.Communication()
	CommunicationCommonInit()
	
	-- Add C4:AddVariables here
end

--[[
	This will be called by the base template code from its OnDriverUpdate function.
--]]
function ON_DRIVER_UPDATE.Communication()
	CommunicationCommonInit()
end

--[[
	This will be called by the base template code from its OnDriverLateInit function.
	
	Available 2.5.0
--]]
function ON_DRIVER_LATEINIT.Communication()

end

function CommunicationCommonInit()
	-- globals	
	gReceiveBuffer = ""
	gCommandQueue = CommandQueue.Create()
	gCommandQueue:SetMaxSize(COMMAND_QUEUE_SIZE)
	gIsNetworkConnected = false		-- Returned by OnNetworkBindingChanged. Indicates if Network connection is bound.
	gIsSerialConnected = false		-- Returned by OnBindingChanged. Indicates if Serial connection is bound.
	
	-- timers
	-- Note the use of properties to set the timer interval parameter.
	-- The same properties are used in gSendTimer:StartTimer and gPollingTimer:StartTimer functions, respectively.
	-- This enables adjustment of these interval times from the Driver Property Tab in Composer.
	gSendTimer = Timer:Create("SendCommand", tonumber(Properties["Command Delay Milliseconds"]), "MILLISECONDS", OnSendTimerExpired)
	gPollingTimer = Timer:Create("PollingTimer", tonumber(Properties["Polling Interval Seconds"]), "SECONDS", OnPollingTimerExpired)	
	
	
	--Persist Data
    if (PersistData == nil) then
	   PersistData = {}
    end

	
end

-- Create Overloaded queue and select type of queue to use.
CommandQueue = {}

--COMMAND_QUEUE_OPTION = 1

function CommandQueue:Create()
	local cq = setmetatable( Queue:Create(), { __index = CommandQueue, __tostring = function(mq) return mq:_tostring() end } )
	
	-- Overloaded function
--[[
	function cq:OnMaxSizeReached()
		--Default: clear queue and push value to the queue. (No need to overload,
		--         use "gCommandQueue = Queue.Create()" instead.

		-- Option 1: Do Nothing, new item is not added to queue
		if (COMMAND_QUEUE_OPTION == 1) then
			Dbg:Info("Max Size Reached - do nothing, new item not added to queue (option 1)")
			return (false)

		-- Option 2: pop value, and push new value on queue
		elseif(COMMAND_QUEUE_OPTION == 2) then
			Dbg:Info("Max Size Reached - pop value, and push new value on queue (option 2)")
			cq:pop()
			return (true)

		-- Option 3: clear queue and DO NOT push new value onto queue
		elseif(COMMAND_QUEUE_OPTION == 3) then
			Dbg:Info("Max Size Reached - clear queue and DO NOT push new value onto queue")
			cq:clear()
			return (false)
		end
	end
--]]	
	return cq
end

--[[
	ReceivedFromNetwork(idBinding, nPort, sData)
		Function which dumps the data received from network (hex format) for inspection via print. It then evaluates the response
		for specific delimiters and extracts the necessary components which are then used to do something.
	Parameters
		idBinding
			Binding ID of the proxy that sent a BindMessage to the DriverWorks driver.
		nPort
			Network Port utilized by the network connection
		sData
			Data received over the serial port
--]]
function ReceivedFromNetwork(idBinding, nPort, sData)
	Dbg:Trace("ReceivedFromNetwork(), idBinding = " .. tostring(idBinding) .. ", nPort = " .. tostring(nPort) .. ", sData = " .. sData)
	--hexdump (sData)

	--device is ONLINE so reset polling counter
	gLastCheckin = 0
	
	-- Add data to receive buffer
	--gReceiveBuffer = gReceiveBuffer .. sData

	-- Parse and handle any data received.
	--ParsePacket()
end

--[[
	HTTP Functions

--]]

function urlPost(url, tData, callback)
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 gNetworkIPAddress = PersistData["BindingAddress"]
		 print("gNetworkIPAddress is nil. Getting Persisted Value.")
	end
	
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 print("gNetworkIPAddress is still nil. Aborting urlPost.")
		 return
	end
	
	url = gNetworkIPAddress .. ":" .. NETWORK_PORT .. url
	local data = JSON:encode(tData)	
	local info = {}
	info["CALLBACK"] = callback
	info["URL"] = url
	info["TICKET"] = C4:urlPost(url, data,{["Content-Type"] = "application/json"})
	table.insert(gGlobalTicketHandlers, info)
end

function urlPut(url, tData, callback)
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 gNetworkIPAddress = PersistData["BindingAddress"]
		 print("gNetworkIPAddress is nil. Getting Persisted Value.")
	end
	
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 print("gNetworkIPAddress is still nil. Aborting urlPut.")
		 return
	end
	
	url = gNetworkIPAddress .. ":" .. NETWORK_PORT .. url
	local data = JSON:encode(tData)
	local info = {}
	info["CALLBACK"] = callback
	info["URL"] = url
	info["TICKET"] = C4:urlPut(url, data,{["Content-Type"] = "application/json"})
	table.insert(gGlobalTicketHandlers, info)
end

function urlGet(url, callback)
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 gNetworkIPAddress = PersistData["BindingAddress"]
		 print("gNetworkIPAddress is nil. Getting Persisted Value.")
	end
	
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 print("gNetworkIPAddress is still nil. Aborting urlGet.")
		 return
	end
	
	url = gNetworkIPAddress .. ":" .. NETWORK_PORT .. url
	local info = {}
	info["CALLBACK"] = callback
	info["URL"] = url
	info["TICKET"] = C4:urlGet(url)
	table.insert(gGlobalTicketHandlers, info)
end

function ReceivedAsync(ticketId, strData, errMsg, responseCode, tHeaders)	
	Dbg:Trace("ReceivedAsync  " .. strData)
	local info = pullTicketById(ticketId)
	
	if (errMsg == 0) then
		if (strData:find("/BeoZone/Zone/ActiveSourceType") ~= nil) then
			Dbg:Trace ("The device received an invalid SetActiveSource command ------  " .. strData)
			if (gSetInputInProgressTimer:TimerStarted()) then
				gRecoverInputSelectionTimer:StartTimer()
			end		
		elseif (strData:find("Couldn't connect to server") ~= nil) then	
			if (gSetInputInProgressTimer:TimerStarted()) then
				gRecoverInputSelectionTimer:StartTimer()
			end			
		elseif (strData:find("/Ping") ~= nil) then
			gPollingTimer:KillTimer()
		else
			Dbg:Trace ("The device received an invalid command ------  " .. strData)
		end
		return
	end
 
	if (tHeaders ~= nil) then
		if (tHeaders:find("Error") ~= nil) then
			Dbg:Trace ("tHeaders has an Error,  return")
			return
		end
	end
	gLastCheckin = 0 
	
	strData = strData or ""
	errMsg = errMsg or "None"
	responseCode = responseCode or 0
	tHeaders = tHeaders or {}
	
	--local info = pullTicketById(ticketId)
	if (info ~= nil) then  
		if (info["CALLBACK"] ~= nil) then
			callUrlCallback(info, errMsg, responseCode, strData, tHeaders)
		end	
	end
end

function pullTicketById(ticketId)
	for i,v in pairs(gGlobalTicketHandlers) do
		if (v["TICKET"] == ticketId) then
			local ret = v
			table.remove(gGlobalTicketHandlers, i)
			return ret
		end
	end
end

function callUrlCallback(info, errMsg, responseCode, strData, tHeaders)
	local data
	data = JSON:decode(strData)
	if (data == nil) then
		Dbg:Warn("ERROR parsing json data: " .. strData)
		return
	end
	info["CALLBACK"](data)
	
	--callback(strError, responseCode, tHeaders, data)
end

function RELEASE(cmd, delay) 
	cmd = cmd or ""
	delay = delay or 10
	
	local url = cmd .. "/Release"
	local tData = ""
	OneShotTimer.Add(delay, "MILLISECONDS", function() urlPost(url, tData, nil) end, "RELEASE") 
end

function ChannelUpDownTimerExpired ()
	gChannelUpDownTimer:KillTimer ()
	RELEASE2 ()
end

--[[
	ReceivedFromSerial(idBinding, sData)
		Function which dumps the data received from serial (hex format) for inspection via print. It then evaluates the response
		for specific delimiters and extracts the necessary components which are then used to do something.
	Parameters
		idBinding
			Binding ID of the proxy that sent a BindMessage to the DriverWorks driver.
		sData
			Data received over the serial port
--]]
function ReceivedFromSerial(idBinding, sData)
	Dbg:Trace("ReceivedFromSerial(), idBinding = " .. tostring(idBinding) .. ", sData = " .. sData)
	--hexdump (sData)

	-- Add data to receive buffer
	gReceiveBuffer = gReceiveBuffer .. sData

	-- Parse and handle any data received.
	ParsePacket()
end

function SendToNetwork(tCommand)
	local sCommand = BuildCommandString(tCommand)
	--Dbg:Trace("SendToNetwork(" .. NETWORK_BINDING_ID .. ", " .. sCommand .. ")")
	if (gIsNetworkConnected) then
		if (gNetworkStatus == "ONLINE") then
			
			--urlPost(url, tData, callback)
			--C4:urlPost(tCommand,{["Content-Type"] = "application/json"})
			--C4:SendToNetwork(NETWORK_BINDING_ID, NETWORK_PORT, sCommand .. "\r") 
			if (gVolumeIsRamping) then 
				gSendTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]))
			else
				gSendTimer:StartTimer(tonumber(Properties["Command Delay Milliseconds"]))
			end			
		else
			CheckNetworkConnectionStatus()
		end
	else 
		Dbg:Warn("Not connected to network. Command not sent.")
	end
end

function BuildCommandString(tCommand)
	--gCommandID = gCommandID + 1
	--tCommand.id = gCommandID
	local command = JSON:encode(tCommand)
	return command
end

function SendToSerial(sCommand)
	Dbg:Trace("SendToSerial(" .. SERIAL_BINDING_ID .. ", " .. sCommand .. ")")
	if (gIsSerialConnected) then
		--[*TEMPLATE REVIEW*]
		-- Commands typically require terminating character(s)
		-- In the example below, a carriage return ("\r\n" - ASCII format) is being appended.
		-- Sometimes a prefix may also be required. If so, it should be added here in the same manner.
		C4:SendToSerial(SERIAL_BINDING_ID, sCommand .. "\r")
		if (gVolumeIsRamping) then 
			gSendTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]))
		else
			gSendTimer:StartTimer(tonumber(Properties["Command Delay Milliseconds"]))
		end	
	else
		Dbg:Warn("Not connected to serial. Command not sent.")
	end
end

function SendToIR(sCommand)
	Dbg:Trace("SendToIR(" .. IR_BINDING_ID .. ", " .. sCommand .. ")")
	if (gControlMethod == "IR") then
		C4:SendIR(IR_BINDING_ID, sCommand)
		gSendTimer:StartTimer(tonumber(Properties["Command Delay Milliseconds"]))
	else
		Dbg:Warn("Not connected to IR. Command not sent.")
	end
end

function OnSendTimerExpired()
	if (gPowerOnDelayTimer:TimerStarted()) then 
		Dbg:Trace("Send Timer expired - Power On Delay In Progress, killing Send Timer and exiting")
		gSendTimer:KillTimer()
		return 
	end
	
	Dbg:Trace("Send Timer expired")
	gSendTimer:KillTimer()

	local cmd = gCommandQueue:pop()
	Dbg:Trace("Send Timer expired - Pop Command:" .. cmd .. ", Send Next Command")

	if (not gCommandQueue:empty()) then
		SendNextCommand()
	end
end

function OnPollingTimerExpired()
	Dbg:Trace("Polling Timer expired")
	gPollingTimer:KillTimer()

	gLastCheckin = gLastCheckin	or 0
	gLastCheckin = gLastCheckin	+	1
	
	if (gLastCheckin == 3)	then
		C4:UpdateProperty("Connected To Network","false")
		Dbg:Trace("Failed to receive poll responses... initiating network recovery mode...")
		C4:ErrorLog(DRIVER_NAME	.. " is	not	responding...	Disconnecting...")
		C4:NetDisconnect(NETWORK_BINDING_ID, NETWORK_PORT)
		gNetworkReconnectTimer:StartTimer()
		return
	elseif (gLastCheckin > 4)	then
		gLastCheckin = 4
	end
	
	
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 gNetworkIPAddress = PersistData["BindingAddress"]
		 print("gNetworkIPAddress is nil. Getting Persisted Value.")
	end
	
	if (gNetworkIPAddress == "" or gNetworkIPAddress == nil) then
		 print("gNetworkIPAddress is still nil. Aborting urlGet.")
		 return
	end
	
	local url = gNetworkIPAddress .. ":" .. NETWORK_PORT .. "/Ping"
	
	
	C4:urlGet(url)	
	
	gPollingTimer:StartTimer(tonumber(Properties["Polling Interval Seconds"]))
	
	if (gPowerState == "ON") then 	--poll for states
		getVolumeLevel()
	end

end

function OnNetworkReconnectTimerExpired()
	Dbg:Trace("OnNetworkReconnectTimerExpired()")
	gNetworkReconnectTimer:KillTimer()
	
	C4:NetConnect(NETWORK_BINDING_ID, NETWORK_PORT)
end

function QueueCommand(sCommand)
	Dbg:Trace("QueueCommand(), sCommand = ")-- .. sCommand)
	Dbg:Debug(gCommandQueue:_tostring())	

	if (sCommand == nil) or (sCommand == "") then
		return
	end

	if (gCommandQueue:empty()) then
		gCommandQueue:push(sCommand)
		SendCommand(sCommand)
	else
		gCommandQueue:push(sCommand)
		CheckNetworkConnectionStatus()
	end
end

function SendCommand(sCommand)
	Dbg:Trace("SendCommand(), sCommand = ")-- .. sCommand)
	if (gControlMethod == "Network") then
		SendToNetwork(sCommand)
	elseif (gControlMethod == "Serial") then
		SendToSerial(sCommand)
	elseif (gControlMethod == "IR") then
		SendToIR(sCommand)		
	end
end

function SendNextCommand()
	local sCommand = gCommandQueue:value()
	Dbg:Trace("SendNextCommand(), sCommand = " .. sCommand)
	if (sCommand == nil or sCommand == "") then
		gSendTimer:KillTimer()
	else
		SendCommand(sCommand)
	end
end

function CheckNetworkConnectionStatus()
	if (gIsNetworkConnected and gNetworkStatus == "OFFLINE") then
		Dbg:Warn("Network status is OFFLINE. Trying to reconnect to the device's Control port...")
		C4:NetDisconnect(NETWORK_BINDING_ID, NETWORK_PORT)
		gNetworkReconnectTimer:StartTimer()
	end
end

-- Receiving Message

function ParsePacket()
	-- Parse and handle any data received.
	local message = GetMessage()
	
	while (message ~= nil and message ~= "") do
		HandleMessage(message)
		
		message = GetMessage()
	end
end

--[[
	GetMessage()
		GetMessage parses the receive buffer to get the next request/response from the receive buffer.
--]]
function GetMessage()
	Dbg:Trace("GetMessage()")

	-- Get a single message from the buffer and return
	
	return message
end

function HandleACK()
	Dbg:Trace("HandleACK()")
	local cmd = gCommandQueue:pop()
	gSendTimer:StartTimer(tonumber(Properties["Command Delay Milliseconds"]))
	if (not gCommandQueue:empty()) then
		SendNextCommand()
	end
end

--[[
	HandleMessage()
		HandleMessage calls the appropriate handler for the request/response received.
--]]
function HandleMessage(message)
	Dbg:Trace("HandleMessage(" .. message .. ")")

	-- parse message and call appropriate handler DEV_MSG[name]
	local tResponse = DecodeMessage(message)
	if (tResponse.type == "notify") or (tResponse.type == "result") then
		local value = tResponse.value 		-- value parsed from message
		if (value == "ACK") then return end	-- we are not processing ACKs
		if (value == "NAK") then return end	-- do we need to process NAKs?
		if (value == "ERR") then return end	-- do we need to process ERRs?
		local name = tResponse.feature 		-- name parsed from Message
		name = name:gsub("%.", "_")			-- remove period character to form valid function names in DEV_MSG
		
		name = name or message
		value = value or ""

		if (DEV_MSG[name] ~= nil and (type(DEV_MSG[name]) == "function")) then
			DEV_MSG[name](value)
		else
			Dbg:Info("HandleMessage: Unhandled message = " .. name)
		end	
	else
		Dbg:Info("HandleMessage: Unhandled type = " .. message)
	end
end

function DecodeMessage(message)
	local tResponse = JSON:decode(message)
	return tResponse
end

--[[
	Define any functions of device messages (DEV_MSG.<name>) received from device that need to be handled by the driver.
--]]

----------------------------------------- Connection Functions -----------------------------------------
--[[
	OnBindingChanged(idBinding, class, bIsBound)
		Function called by Director when a binding changes state (bound or unbound).

	Parameters
		idBinding
			ID of the binding whose state has changed.
		class
			Class of binding that has changed. A single binding can have multiple classes: COMPONENT, STEREO, RS_232, etc.
			This indicates which has been bound or unbound.
		bIsBound
			Whether the binding has been bound or unbound.
--]]
function OnBindingChanged(idBinding, class, bIsBound)
	Dbg:Trace("OnBindingChanged(), idBinding = " .. tostring(idBinding) .. ", class = " .. class .. ", bIsBound = " .. tostring(bIsBound))
	
	if(idBinding == SERIAL_BINDING_ID) then
		gIsSerialConnected = bIsBound
	end	
	if(idBinding == IR_BINDING_ID) then
		gIsIRConnected = bIsBound
	end
	SetControlMethod()
end

--[[
	OnNetworkBindingChanged(idBinding, bIsBound)
		Function called by Director when a network binding changes state (bound or unbound).

	Parameters
		idBinding
			ID of the binding whose state has changed.
		bIsBound
			Whether the binding has been bound or unbound.
--]]
function OnNetworkBindingChanged(idBinding, bIsBound)
	Dbg:Trace('OnNetworkBindingChanged, idBinding = ' .. tostring(idBinding) .. ' bIsBound = ' .. tostring(bIsBound))
	
	--Network binding has changed so reset polling counter
	gLastCheckin = 0
	
	gIsNetworkConnected = bIsBound
	if (bIsBound == false) then
		gPollingTimer:KillTimer()
	end

	SetControlMethod()
end

function SetControlMethod()
	if (gIsNetworkConnected) then
		gControlMethod = "Network"
	elseif (gIsSerialConnected) then
		gControlMethod = "Serial"	
	elseif (gIsIRConnected) then
		gControlMethod = "IR"		
	else	
		gControlMethod = "(none)"	
	end
	gCommandQueue:clear()
	return gControlMethod
end

function ValidateControlMethod(control_method)
	local isValid = false
	if (control_method=="Network") and (gIsNetworkConnected) then
		isValid = true
	elseif (control_method=="Serial") and (gIsSerialConnected) then
		isValid = true	
	end
	return isValid
end

function OnConnectionStatusChanged(idBinding, nPort, sStatus)
	Dbg:Trace("OnConnectionStatusChanged[" .. idBinding .. " (" .. nPort .. ")]: " .. sStatus)

	if (idBinding == NETWORK_BINDING_ID) then
		gNetworkStatus=sStatus
		if (sStatus == "ONLINE") then
			gNetworkIPAddress = C4:GetMyNetworkAddress()
			
			if (gNetworkIPAddress ~= "" and gNetworkIPAddress ~= nil) then
				PersistData["BindingAddress"] = gNetworkIPAddress
			end
			
			local next = next
			
			--moving to ActiveSourceType command 
			--[[
			if (next(tInputCommandMap) == nil) then
				get_input_selection_commands() 
			end
			--]]
			
			if (gPowerState == "" or gPowerState == nil) then
				get_power_status()
			end
			
			if (next(tPowerCommandMap) == nil) then
				get_power_commands()
			end	
			
			if (next(tSoundModeCommandMap) == nil) then
				get_sound_mode_commands()
			end	
			
			if (gIsTV) then
				get_mac_address()
				
				if (gSetInputInProgressTimer:TimerStarted()) then
					gSetInputInProgressTimer:KillTimer()
					gRecoverInputSelectionTimer:StartTimer()
				end

				if (next(tStandCommandMap) == nil) then
					get_stand_commands()
				end	
				
				if (next(tSpeakerGroupCommandMap) == nil) then
					get_speaker_group_commands()
				end	
				if (next(tPictureModeCommandMap) == nil) then
					get_picture_mode_commands()	
				end		

				if (next(tPictureFormat2d3dCommandMap) == nil) then
					get_picture_format_2d3d_commands()	
				end	
			end

			if (gPollingTimer:TimerStopped()) then
				gPollingTimer:StartTimer(tonumber(Properties["Polling Interval Seconds"]))
			end
			
			--device is ONLINE so reset polling counter
			gLastCheckin = 0
			C4:UpdateProperty("Connected To Network","true")
		end				 
	else
		gPollingTimer:KillTimer()
	end
end

--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Debug Logging Code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
Log = {}

-- Create a Table with Logging functions
function Log:Create()
	
	-- table for logging functions
	local lt = {}
	
	lt._logLevel = 0
	lt._outputPrint = false
	lt._outputC4Log = false
	lt._logName =  "Set Log Name to display"
	
	function lt:SetLogLevel(level)
		self._logLevel = level
	end
	
	function lt:OutputPrint(value)
		self._outputPrint = value
	end
	
	function lt:OutputC4Log(value)
		self._outputC4Log = value
	end
	
	function lt:SetLogName(name)
		self._logName = name
	end

	function lt:Enabled()
		return (self._outputPrint or self._outputC4Log)
	end
	
	function lt:PrintTable(tValue, sIndent)
		if (type(tValue) == "table") then
			if (self._outputPrint) then
				for k,v in pairs(tValue) do
					print(sIndent .. tostring(k) .. ":  " .. tostring(v))
					if (type(v) == "table") then
						self:PrintTable(v, sIndent .. "   ")
					end
				end
			end
	
			if (self._outputC4Log) then
				for k,v in pairs(tValue) do
					C4:ErrorLog(self._logName .. ": " .. sIndent .. tostring(k) .. ":  " .. tostring(v))
					if (type(v) == "table") then
						self:PrintTable(v, sIndent .. "   ")
					end
				end
			end

		else
			if (self._outputPrint) then
				print (sIndent .. tValue)
			end
			
			if (self._outputC4Log) then
				C4:ErrorLog(self._logName .. ": " .. sIndent .. tValue)
			end
		end
	end
		
	function lt:Print(logLevel, sLogText)
		if (self._logLevel >= logLevel) then
			if (type(sLogText) == "table") then
				self:PrintTable(sLogText, "   ")
				return
			end
			
			if (self._outputPrint) then
				print (sLogText)
			end

			if (self._outputC4Log) then
				C4:ErrorLog(self._logName .. ": " .. sLogText)
			end
		end
	end
	
	function lt:Alert(strDebugText)
		self:Print(0, strDebugText)
	end
	
	function lt:Error(strDebugText)
		self:Print(1, strDebugText)
	end
	
	function lt:Warn(strDebugText)
		self:Print(2, strDebugText)
	end
	
	function lt:Info(strDebugText)
		self:Print(3, strDebugText)
	end
	
	function lt:Trace(strDebugText)
		self:Print(4, strDebugText)
	end
	
	function lt:Debug(strDebugText)
		self:Print(5, strDebugText)
	end
	
	return lt
end

function ON_DRIVER_EARLY_INIT.LogLib()
	-- Create and initialize debug logging
	DRIVER_NAME = C4:GetDriverConfigInfo("name")
	
	Dbg = Log.Create()
	Dbg:SetLogName(DRIVER_NAME)
end

function ON_DRIVER_INIT.LogLib()
	-- Create Debug Timer
	gDebugTimer = Timer:Create("Debug", 45, "MINUTES", OnDebugTimerExpired)
end

--[[
	OnDebugTimerExpired
		Debug timer callback function
--]]
function OnDebugTimerExpired()
	Dbg:Warn("Turning Debug Mode Off (timer expired)")
	gDebugTimer:KillTimer()
	C4:UpdateProperty("Debug Mode", "Off")
	OnPropertyChanged("Debug Mode")
end
---------------------------------------------------------------------
-- Timer Code
---------------------------------------------------------------------
Timer = {}

-- Create a Table with Timer functions
function Timer:Create(name, interval, units, Callback, repeating, Info)
	-- timers table
	local tt = {}
	
	tt._name = name
	tt._timerID = TimerLibGetNextTimerID()
	tt._interval = interval
	tt._units = units
	tt._repeating = repeating or false
	tt._Callback = Callback
	tt._info = Info or ""
	tt._id = 0

	function tt:StartTimer(...)
		self:KillTimer()
		
		-- optional parameters (interval, units, repeating)
		if ... then
			local interval = select(1, ...)
			local units = select(2, ...)
			local repeating = select(3, ...)
			
			self._interval = interval or self._interval
			self._units = units or self._units
			self._repeating = repeating or self._repeating
		end
		
		if (self._interval > 0) then
			Dbg:Trace("Starting Timer: " .. self._name)
			self._id = C4:AddTimer(self._interval, self._units, self._repeating)
		end
	end

	function tt:KillTimer()
		if (self._id) then
			self._id = C4:KillTimer(self._id)
		end
	end
	
	function tt:TimerStarted()
		return (self._id ~= 0)
	end
			
	function tt:TimerStopped()
		return not self:TimerStarted()
	end
	
	gTimerLibTimers[tt._timerID] = tt
	Dbg:Trace("Created timer " .. tt._name)
	
	return tt
end

function TimerLibGetNextTimerID()
	gTimerLibTimerCurID = gTimerLibTimerCurID + 1
	return gTimerLibTimerCurID
end

function ON_DRIVER_EARLY_INIT.TimerLib()
	gTimerLibTimers = {}
	gTimerLibTimerCurID = 0
end

function ON_DRIVER_DESTROYED.TimerLib()
	-- Kill open timers
	for k,v in pairs(gTimerLibTimers) do
		v:KillTimer()
	end
end

--[[
	OnTimerExpired
		Function called by Director when the specified Control4 timer expires.
	Parameters
		idTimer
			Timer ID of expired timer.
--]]
function OnTimerExpired(idTimer)
	for k,v in pairs(gTimerLibTimers) do
		if (idTimer == v._id) then
			if (v._Callback) then
				v._Callback(v._info)
			end
		end
	end
end
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Queue Code
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
Queue = {}

-- Create a Table with Queue functions
function Queue:Create()

	-- queue table
	local t = setmetatable(Queue, {
		__tostring = function(t)
			return t:_tostring()
		end
	})

	function t:_tostring()
		local tOutputString = {}
		
		table.insert(tOutputString, "--- Queue ---")
		table.insert(tOutputString, "  first = " .. tostring(self._et.first))
		table.insert(tOutputString, "  last = " .. tostring(self._et.last))
		table.insert(tOutputString, "  number in queue = " .. tostring(self._et.last - self._et.first + 1))
		table.insert(tOutputString, "  maximum size = " .. self._maxSize)
		table.insert(tOutputString, "  next value = " .. self:value())
		return table.concat(tOutputString, "\n")
	end

	-- entry table
	t._et = {first = 0, last = -1}
	t._maxSize = 0	-- no size limit

	-- push a value on the queue
	function t:push (value)
		local numItems = self._et.last - self._et.first + 1

		if ( (self._maxSize <= 0) or (numItems < self._maxSize) ) then
			local last = self._et.last + 1
			self._et.last = last
			self._et[last] = value
			--Dbg:Trace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		else
			-- if addToQueue == true then push value to queue
			if (self.OnMaxSizeReached()) then
				local last = self._et.last + 1
				self._et.last = last
				self._et[last] = value
				--Dbg:Trace ("Queue:push(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
			end
		end
	end

	function t:OnMaxSizeReached()
		--Dbg:Trace ("Max Size Reached - clear queue and push value to the queue (default).")
		local addToQueue = true

		t:clear()
		return (addToQueue)
	end

	-- pop a value from the queue
	function t:pop ()
		local first = self._et.first
		if first > self._et.last then
			--Dbg:Trace("Queue:pop(), queue is empty")
			return ""
		end
		local value = self._et[first]
		self._et[first] = nil        -- to allow garbage collection
		self._et.first = first + 1
		--Dbg:Trace ("Queue:pop(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1) .. ", value = " .. value)
		return value
	end

	-- clear queue
	function t:clear ()
		local first = self._et.first
		if first > self._et.last then
			--Dbg:Trace ("Queue:clear(), queue is empty")
			return ""
		end
		
		self._et = {first = 0, last = -1}
		--Dbg:Trace ("Queue:clear(), first = " .. tostring(self._et.first) .. ", last = " .. tostring(self._et.last) .. ", number in queue = " .. tostring(self._et.last - self._et.first + 1))
		--Dbg:Trace (self._et)
	end

	-- return value of first item
	function t:value ()
		local first = self._et.first
		if (first > self._et.last) then
			return ""
		else
			return self._et[first]
		end
	end

	-- return queue's maximum size
	function t:MaxSize ()
		return self._maxSize
	end

	-- return queue's maximum size
	function t:SetMaxSize (size)
		self._maxSize = size
	end

	-- is queue empty?
	function t:empty ()
		-- print ("self._et.first = " .. tostring(self._et.first) .. ", self._et.last = " .. tostring(self._et.last))
		if (self._et.first > self._et.last) then
			return true
		else
			return false
		end
	end
	
	return t
end
function SendToProxy(idBinding, strCommand, tParams)
	Dbg:Trace("SendToProxy(" .. idBinding .. ", " .. strCommand .. ")")
	Dbg:Trace(tParams)
	
	C4:SendToProxy(idBinding, strCommand, tParams)
end

function ReverseTable(a)
	local b = {}
	for k,v in pairs(a) do b[v] = k end
	return b
end


--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- TV Protocol Declarations
--=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=
-- Constants
TV_PROXY_BINDINGID = 5001
DEEZER_PROXY_BINDINGID = 5002
HOMEMEDIA_PROXY_BINDINGID = 5003
SPOTIFY_PROXY_BINDINGID = 5005
WEBMEDIA_PROXY_BINDINGID = 5006
YOUTUBE_PROXY_BINDINGID = 5007
NETRADIO_PROXY_BINDINGID = 5008
RADIO_PROXY_BINDINGID = 5011
RECORDINGS_PROXY_BINDINGID = 5012
DVBRADIO_PROXY_BINDINGID = 5013
SAT_PROXY_BINDINGID = 5014

PROTOCOL_DECLARATIONS = {}

function ON_DRIVER_INIT.ProtocolDeclarations()
	Dbg:Trace("ON_DRIVER_INIT.ProtocolDeclarations()")

	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end	
	
end

function ON_DRIVER_UPDATE.ProtocolDeclarations()
	Dbg:Trace("ON_DRIVER_UPDATE.ProtocolDeclarations()")
	
	for k,v in pairs(PROTOCOL_DECLARATIONS) do
		if (PROTOCOL_DECLARATIONS[k] ~= nil and type(PROTOCOL_DECLARATIONS[k]) == "function") then
			PROTOCOL_DECLARATIONS[k]()
		end
	end	
	
end

function PROTOCOL_DECLARATIONS.SetGlobals()
	Dbg:Trace("PROTOCOL_DECLARATIONS.SetGlobals()")
	
	--load JSON library
	require "json"
	JSON=(loadstring(json.JSON_LIBRARY_CHUNK))()
	
	gGlobalTicketHandlers = {}
	
	--Network variable
	gNetworkIPAddress = ""
	
	-- Status Variables
	gPowerState = ""
	
	-- Volume Handling
	-- Mode Values: VOLUME_UP & VOLME_DOWN
	gVolumeIsRamping = false
	gVolumeRamping = {
		[0] = {state = false,mode = "",},
	}	
	gCurrentVolume = {}
	gLastVolumeStatusValue = {}		
	
	gMuteState = {}
	
	--timers
	gPowerOnDelayTimer = Timer:Create("PowerOnDelayTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnDelayTimerExpired)
	gRecoverInputSelectionTimer = Timer:Create("RecoverInputSelectionTimer", 2, "SECONDS", OnRecoverInputSelectionTimerExpired)
	gSetInputInProgressTimer = Timer:Create("SetInputInProgressTimer", 60, "SECONDS", OnSetInputInProgressTimerExpired)
	gChannelUpDownTimer = Timer:Create ("ChannelUpDownTimer", 45, "SECONDS", ChannelUpDownTimerExpired)
	gNetworkReconnectTimer = Timer:Create ("NetworkReconnectTimer", 5, "SECONDS", OnNetworkReconnectTimerExpired)

	gIsTV = isTV()
	
end

function isTV()
  local bTV = false
  if (string.find(C4:GetDriverConfigInfo ("proxies"), "tv")) then  bTV = true end
  return bTV
end
	
function PROTOCOL_DECLARATIONS.CommandsTableInit_IR()
	Dbg:Trace("PROTOCOL_DECLARATIONS.CommandsTableInit_IR()")

	CMDS_IR = {
		--index:  Proxy Command Name
		--value:  IR Code ID	 
		
		--Power
		["ON"]             = "",
		["OFF"]            = "",
		
		--Input
		['INPUT_TOGGLE']       = '',
				
		--Channel
		['PULSE_CH_DOWN']      = '',
		['PULSE_CH_UP']        = '',
		['START_CH_DOWN']      = '',
		['START_CH_UP']        = '',		
		['STOP_CH_DOWN']       = '',
		['STOP_CH_UP']         = '',
		
		--Volume		
		['PULSE_VOL_UP']       = '',
		['PUSLE_VOL_DOWN']     = '',
		['MUTE_OFF']           = '',
		['MUTE_ON']            = '',
		['MUTE_TOGGLE']        = '',
		
		--Menu
		["INFO"] = "",	--Display
		["GUIDE"] = "",
		["MENU"] = "",
		["CANCEL"] = "",
		["UP"] = "",
		["DOWN"] = "",
		["LEFT"] = "",
		["RIGHT"] = "",
		["ENTER"] = "",	
		["RECALL"]         = "",
		["PREV"]           = "",
		['PAGE_DOWN']          = '',
		['PAGE_UP']            = '',		
		
		--Digits
		["NUMBER_0"]       = "", 	
		["NUMBER_1"]       = "",		
		["NUMBER_2"]       = "",	
		["NUMBER_3"]       = "",	
		["NUMBER_4"]       = "",	
		["NUMBER_5"]       = "",	
		["NUMBER_6"]       = "",	
		["NUMBER_7"]       = "",	
		["NUMBER_8"]       = "",	
		["NUMBER_9"]       = "",
		["STAR"]           = "",
		["DOT"]            = "",
		
		--Transport
		["OPEN"]			= "",
		["CLOSE"]			= "",
		["OPEN_CLOSE"]		= "",
		["PAUSE"]			= "",
		["PLAY"]			= "",
		["SCAN_FWD"]		= "",
		["SCAN_REV"]		= "",
		["SKIP_FWD"]		= "",
		["SKIP_REV"]		= "",
		["STOP"]			= "",	
		
		--Program / Color Buttons
		["PROGRAM_A"]		= "",
		["PROGRAM_B"]		= "",
		["PROGRAM_D"]		= "",
		["PROGRAM_C"]		= "",		
	}		
		
end

function PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()
	Dbg:Trace("PROTOCOL_DECLARATIONS.CommandsTableInit_Serial()")

	CMDS = {
		--index:  Proxy Command Name
		--value:  Protocol Command Data
		
		--Power
		["ON"]             = "",
		["OFF"]            = "",
		
		--Input
		['INPUT_TOGGLE']       = '',
		
		--Channel
		['PULSE_CH_DOWN']      = '',
		['PULSE_CH_UP']        = '',
		['START_CH_DOWN']      = '',
		['START_CH_UP']        = '',		
		['STOP_CH_DOWN']       = '',
		['STOP_CH_UP']         = '',
		
		--Volume		
		['PULSE_VOL_UP']       = '',
		['PUSLE_VOL_DOWN']     = '',
		['MUTE_OFF']           = '',
		['MUTE_ON']            = '',
		['MUTE_TOGGLE']        = '',
		
		--Menu
		["INFO"] = "",	--Display
		["GUIDE"] = "",
		["MENU"] = "",
		["CANCEL"] = "",
		["UP"] = "",
		["DOWN"] = "",
		["LEFT"] = "",
		["RIGHT"] = "",
		["ENTER"] = "",	
		["RECALL"]         = "",
		["PREV"]           = "",
		['PAGE_DOWN']          = '',
		['PAGE_UP']            = '',			
		
		--Digits
		["NUMBER_0"]       = "", 	
		["NUMBER_1"]       = "",		
		["NUMBER_2"]       = "",	
		["NUMBER_3"]       = "",	
		["NUMBER_4"]       = "",	
		["NUMBER_5"]       = "",	
		["NUMBER_6"]       = "",	
		["NUMBER_7"]       = "",	
		["NUMBER_8"]       = "",	
		["NUMBER_9"]       = "",
		["STAR"]           = "",
		["DOT"]            = "",
		
		--Transport
		["OPEN"]			= "",
		["CLOSE"]			= "",
		["OPEN_CLOSE"]		= "",
		["PAUSE"]			= "",
		["PLAY"]			= "",
		["SCAN_FWD"]		= "",
		["SCAN_REV"]		= "",
		["SKIP_FWD"]		= "",
		["SKIP_REV"]		= "",
		["STOP"]			= "",	
		
		--Program / Color Buttons
		["PROGRAM_A"]		= "",
		["PROGRAM_B"]		= "",
		["PROGRAM_D"]		= "",
		["PROGRAM_C"]		= "",		
	}
	
end

function PROTOCOL_DECLARATIONS.InputOutputTableInit()
	Dbg:Trace("PROTOCOL_DECLARATIONS.InputOutputTableInit()")
	----------------------------------------- [*COMMAND/RESPONSE HELPER TABLES*] -----------------------------------------
	
	--Persist Data
	--[[
    if (PersistData == nil) then
	   PersistData = {}
    end

    if(PersistData["InputCommandMap"] == nil) then
	   PersistData["InputCommandMap"] = {}
    end    
    tInputCommandMap = PersistData["InputCommandMap"]
    
    if(PersistData["PowerCommandMap"] == nil) then
	   PersistData["PowerCommandMap"] = {}
    end    
    tPowerCommandMap = PersistData["PowerCommandMap"]   
	--]]
	
	--tInputCommandMap = {}
	tPowerCommandMap = {}
	tStandCommandMap = {}
	tSpeakerGroupCommandMap = {}
	tPictureModeCommandMap = {}
	tSoundModeCommandMap = {}
	tPictureFormat2d3dCommandMap = {}
	
	tInputCommandMap = { -- this is empty because it is built dynamically
		--index:  Connection Name
		--value:  Protocol Command Data	tInputCommandMap[tInputConnMapByID[input].Name]
		["tv"] = {sourceType = {["type"] = "TV"}}, 
		["hdmi_1"] = {sourceType = {["type"] = "HDMI", ["connector"] = 1}},
		["hdmi_2"] = {sourceType = {["type"] = "HDMI", ["connector"] = 2}},		
		["hdmi_3"] = {sourceType = {["type"] = "HDMI", ["connector"] = 3}},
		["hdmi_4"] = {sourceType = {["type"] = "HDMI", ["connector"] = 4}},
		["hdmi_5"] = {sourceType = {["type"] = "HDMI", ["connector"] = 5}},
		["hdmi_6"] = {sourceType = {["type"] = "HDMI", ["connector"] = 6}}, 
		["hdmi_matrix_1"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 1}},
		["hdmi_matrix_2"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 2}},		
		["hdmi_matrix_3"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 3}},		
		["hdmi_matrix_4"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 4}},
		["hdmi_matrix_5"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 5}},
		["hdmi_matrix_6"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 6}}, 
		["hdmi_matrix_7"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 7}},
		["hdmi_matrix_8"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 8}},		
		["hdmi_matrix_9"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 9}},		
		["hdmi_matrix_10"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 10}},
		["hdmi_matrix_11"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 11}},
		["hdmi_matrix_12"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 12}}, 
		["hdmi_matrix_13"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 13}},
		["hdmi_matrix_14"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 14}},		
		["hdmi_matrix_15"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 15}},		
		["hdmi_matrix_16"] = {sourceType = {["type"] = "MATRIX", ["connector"] = 16}},
		["av"] = {sourceType = {["type"] = "AV IN"}}, 
		["linein"] = {sourceType = {["type"] = "LINE IN"}},
		["CD"] = {sourceType = {["type"] = "CD"}},		
		["dvb_radio"] = {sourceType = {["type"] = "DVB_RADIO"}},		
		--["TP1"] = {sourceType = {["type"] = ""}},
		["AUX_A"] = {sourceType = {["type"] = "A.AUX"}}, 
		--["TP2"] = {sourceType = {["type"] = "AV IN"}},
		["homemedia"] = {sourceType = {["type"] = "HOMEMEDIA"}},		
		["webmedia"] = {sourceType = {["type"] = "WEBMEDIA"}},
		["netradio"] = {sourceType = {["type"] = "NET RADIO"}},		
		["youtube"] = {sourceType = {["type"] = "YOUTUBE"}},
		["spotify"] = {sourceType = {["type"] = "SPOTIFY"}},		
		["deezer"] = {sourceType = {["type"] = "DEEZER"}},
		--["PH"] = {sourceType = {["type"] = ""}},
		["tunein"] = {sourceType = {["type"] = "TUNEIN"}},
		["radio"] = {sourceType = {["type"] = "RADIO"}},
		["music"] = {sourceType = {["type"] = "MUSIC"}},
		["toslink"] = {sourceType = {["type"] = "TOS LINK"}},
		["recordings"] = {sourceType = {["type"] = "RECORDINGS"}},
		["sat"] = {sourceType = {["type"] = "SAT"}},
		
	}

	tInputResponseMap = {
		["TV"] = "tv",
		["HDMI1"] = "hdmi_1", 	
		["HDMI2"] = "hdmi_2",		
		["HDMI3"] = "hdmi_3",
		["HDMI4"] = "hdmi_4",
		["HDMI5"] = "hdmi_5",
		["HDMI6"] = "hdmi_6",
		["MATRIX1"] = "hdmi_matrix_1",	
		["MATRIX2"] = "hdmi_matrix_2",	
		["MATRIX3"] = "hdmi_matrix_3",	
		["MATRIX4"] = "hdmi_matrix_4",	
		["MATRIX5"] = "hdmi_matrix_5",	
		["MATRIX6"] = "hdmi_matrix_6",	
		["MATRIX7"] = "hdmi_matrix_7",	
		["MATRIX8"] = "hdmi_matrix_8",	
		["MATRIX9"] = "hdmi_matrix_9",	
		["MATRIX10"] = "hdmi_matrix_10",	
		["MATRIX11"] = "hdmi_matrix_11",	
		["MATRIX12"] = "hdmi_matrix_12",	
		["MATRIX13"] = "hdmi_matrix_13",	
		["MATRIX14"] = "hdmi_matrix_14",	
		["MATRIX15"] = "hdmi_matrix_15",	
		["MATRIX16"] = "hdmi_matrix_16",	
		["AV IN"] = "av",
		["LINE IN"] = "linein",
		["CD"] = "CD",
		["DVB_RADIO"] = "dvb_radio",
		["A.AUX"] = "AUX_A",
		["HOMEMEDIA"] = "homemedia",
		["WEBMEDIA"] = "webmedia",
		["NET RADIO"] = "netradio",
		["YOUTUBE"] = "youtube",
		["SPOTIFY"] = "spotify",
		["DEEZER"] = "deezer",
		["TUNEIN"] = "tunein",
		["RADIO"] = "radio",
		["MUSIC"] = "music",
		["TOS LINK"] = "toslink",
		["RECORDINGS"] = "recordings",
		["SAT"] = "sat",
	}

	tInputConnMapByID = {
		--index:  mod 1000 value of Input Connection id
		--[0] = {Name = "INPUT HDMI 1",BindingID = TV_PROXY_BINDINGID,},
		[0] = {Name = "tv",BindingID = TV_PROXY_BINDINGID,},
		[1] = {Name = "hdmi_1",BindingID = TV_PROXY_BINDINGID,},
		[2] = {Name = "hdmi_2",BindingID = TV_PROXY_BINDINGID,},		
		[3] = {Name = "hdmi_3",BindingID = TV_PROXY_BINDINGID,},
		[4] = {Name = "hdmi_4",BindingID = TV_PROXY_BINDINGID,},
		[5] = {Name = "hdmi_5",BindingID = TV_PROXY_BINDINGID,},
		[6] = {Name = "hdmi_6",BindingID = TV_PROXY_BINDINGID,},
		[7] = {Name = "hdmi_matrix_1",BindingID = TV_PROXY_BINDINGID,},
		[8] = {Name = "hdmi_matrix_2",BindingID = TV_PROXY_BINDINGID,},		
		[9] = {Name = "hdmi_matrix_3",BindingID = TV_PROXY_BINDINGID,},
		[10] = {Name = "hdmi_matrix_4",BindingID = TV_PROXY_BINDINGID,},
		[11] = {Name = "hdmi_matrix_5 ",BindingID = TV_PROXY_BINDINGID,},
		[12] = {Name = "hdmi_matrix_6",BindingID = TV_PROXY_BINDINGID,},
		[13] = {Name = "hdmi_matrix_7",BindingID = TV_PROXY_BINDINGID,},
		[14] = {Name = "hdmi_matrix_8",BindingID = TV_PROXY_BINDINGID,},		
		[15] = {Name = "hdmi_matrix_9",BindingID = TV_PROXY_BINDINGID,},
		[16] = {Name = "hdmi_matrix_10",BindingID = TV_PROXY_BINDINGID,},
		[17] = {Name = "hdmi_matrix_11",BindingID = TV_PROXY_BINDINGID,},
		[18] = {Name = "hdmi_matrix_12",BindingID = TV_PROXY_BINDINGID,},
		[19] = {Name = "hdmi_matrix_13",BindingID = TV_PROXY_BINDINGID,},
		[20] = {Name = "hdmi_matrix_14",BindingID = TV_PROXY_BINDINGID,},		
		[21] = {Name = "hdmi_matrix_15",BindingID = TV_PROXY_BINDINGID,},
		[22] = {Name = "hdmi_matrix_16",BindingID = TV_PROXY_BINDINGID,},
		[23] = {Name = "av",BindingID = TV_PROXY_BINDINGID,},
		[24] = {Name = "linein",BindingID = TV_PROXY_BINDINGID,},
		[25] = {Name = "CD",BindingID = TV_PROXY_BINDINGID,},
		[26] = {Name = "dvb_radio",BindingID = TV_PROXY_BINDINGID,},
		[27] = {Name = "TP1",BindingID = TV_PROXY_BINDINGID,},
		[28] = {Name = "AUX_A",BindingID = TV_PROXY_BINDINGID,},
		[29] = {Name = "TP2",BindingID = TV_PROXY_BINDINGID,},
		[31] = {Name = "homemedia",BindingID = TV_PROXY_BINDINGID,},
		[34] = {Name = "webmedia",BindingID = TV_PROXY_BINDINGID,},
		[32] = {Name = "youtube",BindingID = TV_PROXY_BINDINGID,},
		[33] = {Name = "spotify",BindingID = TV_PROXY_BINDINGID,},
		[30] = {Name = "deezer",BindingID = TV_PROXY_BINDINGID,},	
		[35] = {Name = "netradio",BindingID = TV_PROXY_BINDINGID,},
		[36] = {Name = "PH",BindingID = TV_PROXY_BINDINGID,},
		[37] = {Name = "tunein",BindingID = TV_PROXY_BINDINGID,},	--radio
		[38] = {Name = "music",BindingID = TV_PROXY_BINDINGID,},
		[40] = {Name = "toslink",BindingID = TV_PROXY_BINDINGID,},
		[41] = {Name = "radio",BindingID = TV_PROXY_BINDINGID,},
		[42] = {Name = "recordings",BindingID = TV_PROXY_BINDINGID,},
		[43] = {Name = "dvb_radio",BindingID = TV_PROXY_BINDINGID,},
		[44] = {Name = "sat",BindingID = TV_PROXY_BINDINGID,},
		
	}
	
	tInputConnMapByName = {
		--index:  Input Connection Name
		--ID: mod 1000 value of Input Connection id
		--["INPUT HDMI 1"] = {ID = 0,BindingID = AVSWITCH_PROXY_BINDINGID,},
		["tv"] = {ID = 0,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_1"] = {ID = 1,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_2"] = {ID = 2,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_3"] = {ID = 3,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_4"] = {ID = 4,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_5"] = {ID = 5,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_6"] = {ID = 6,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_1"] = {ID = 7,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_2"] = {ID = 8,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_3"] = {ID = 9,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_4"] = {ID = 10,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_5"] = {ID = 11,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_6"] = {ID = 12,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_7"] = {ID = 13,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_8"] = {ID = 14,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_9"] = {ID = 15,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_10"] = {ID = 16,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_11"] = {ID = 17,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_12"] = {ID = 18,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_13"] = {ID = 19,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_14"] = {ID = 20,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_15"] = {ID = 21,BindingID = TV_PROXY_BINDINGID,},
		["hdmi_matrix_16"] = {ID = 22,BindingID = TV_PROXY_BINDINGID,},
		["av"] = {ID = 23,BindingID = TV_PROXY_BINDINGID,},
		["linein"] = {ID = 24,BindingID = TV_PROXY_BINDINGID,},
		["CD"] = {ID = 25,BindingID = TV_PROXY_BINDINGID,},
		["dvb_radio"] = {ID = 26,BindingID = TV_PROXY_BINDINGID,},
		["TP1"] = {ID = 27,BindingID = TV_PROXY_BINDINGID,},
		["AUX_A"] = {ID = 28,BindingID = TV_PROXY_BINDINGID,},
		["TP2"] = {ID = 29,BindingID = TV_PROXY_BINDINGID,},
		["homemedia"] = {ID = 31,BindingID = TV_PROXY_BINDINGID,},
		["webmedia"] = {ID = 34,BindingID = TV_PROXY_BINDINGID,},
		["youtube"] = {ID = 32,BindingID = TV_PROXY_BINDINGID,},
		["spotify"] = {ID = 33,BindingID = TV_PROXY_BINDINGID,},
		["deezer"] = {ID = 30,BindingID = TV_PROXY_BINDINGID,},
		["netradio"] = {ID = 35,BindingID = TV_PROXY_BINDINGID,},
		["PH"] = {ID = 36,BindingID = TV_PROXY_BINDINGID,},
		["tunein"] = {ID = 37,BindingID = TV_PROXY_BINDINGID,},		--radio
		["music"] = {ID = 38,BindingID = TV_PROXY_BINDINGID,},
		["toslink"] = {ID = 40,BindingID = TV_PROXY_BINDINGID,},
		["radio"] = {ID = 41,BindingID = TV_PROXY_BINDINGID,},
		["recordings"] = {ID = 42,BindingID = TV_PROXY_BINDINGID,},
		["dvb_radio"] = {ID = 43,BindingID = TV_PROXY_BINDINGID,},
		["sat"] = {ID = 44,BindingID = TV_PROXY_BINDINGID,},
	}	
end

----------------------------------------- Device Message Functions -----------------------------------------
--[*TEMPLATE REVIEW*]
--These DEV_MSG functions are called from the HandleMessage() function.
--The function naming syntax is DEV_MSG.NAME where NAME is the portion 
--of the device message that indicate the message type.
--Below are example functions for each proxy notification.
--You will need to edit the portion of the function name after the period 
--to match the actual message that is returned from the device 
--(for example, DEV_MSG.NAME to DEV_MSG.MV if the devices returns the string ?MV? 
--to indicate a volume status message).

function DEV_MSG.POWER(value)
	Dbg:Trace("DEV_MSG.POWER()")
	if (value.standby == nil) then Dbg:Trace("DEV_MSG.POWER(), exiting, value.standby is nil...") return end
	local powerState = value.standby.powerState
	if (powerState == "standby" or powerState == "allStandby") then  --Off State
		gPowerState = "OFF"
		NOTIFY.OFF()
	elseif (powerState == "on") then  
		gPowerState = "ON"
		NOTIFY.ON()
	else
		Dbg:Trace("DEV_MSG.POWER: unknown power state = " .. powerState)
	end
end

function DEV_MSG.INPUT(value)
	if (value.sourceType == nil) then
		Dbg:Trace("DEV_MSG.INPUT(), value.sourceType == nil")
		gRecoverInputSelectionTimer:StartTimer()
		return
	else
		Dbg:Trace("DEV_MSG.INPUT()")
		gLastSelectedInput = ""
		gSetInputInProgressTimer:KillTimer()
		gRecoverInputSelectionTimer:KillTimer()
	end

	local source_type = value.sourceType.type
	local connector = value.sourceType.connector
	if (source_type == nil) then Dbg:Trace("DEV_MSG.INPUT(), exiting, source_type is nil...") return end
	
	local source = ""
	if (source_type == "HDMI" or source_type == "MATRIX") then
		source = source_type .. connector
	else
		source = source_type
	end
	
	local input_name = tInputResponseMap[source]
	local id = tInputConnMapByName[input_name].ID

	Dbg:Trace ("DEV_MSG.INPUT Input Source Name is  =  ".. input_name)
	
	--hdmi & hdmi matrix do not have audio connection
	--if (source_type ~= "HDMI" and source_type == "MATRIX") then
		--NOTIFY.INPUT_CHANGED(id + 3000)
	--end		
	
	--TV proxy bug, ALL connections need audio notification, even HDMI only connections...
	NOTIFY.INPUT_CHANGED(id + 3000)
	if (gIsTV) then
		--tv (not receiver) must report video path
		NOTIFY.INPUT_CHANGED(id + 1000)
	end
	
end

function DEV_MSG.CHANNEL(value)
	Dbg:Trace("DEV_MSG.CHANNEL()")									
	NOTIFY.CHANNEL_CHANGED(value)
end

function DEV_MSG.VOLUME(value)
	 Dbg:Trace("DEV_MSG.VOLUME()")

	 local device_level = value.level
	 local output = 0 --mod 1000 value of Output Connection ID
	 
	 --Edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	 local minDeviceLevel = 0
	 local maxDeviceLevel = 100
	 --C4 volume level is represented as percentage value
	 local c4_level = ConvertVolumeToC4(device_level, minDeviceLevel, maxDeviceLevel)                    
	 
	 if (gLastVolumeStatusValue[output] ~= device_level) then
		NOTIFY.VOLUME_LEVEL_CHANGED(output, c4_level)
	 end
	 gLastVolumeStatusValue[output] = device_level

	 if (gVolumeRamping[output].state == true) then
		ContinueVolumeRamping()    
	 else
		local url = "/BeoZone/Zone/Sound/Volume/Speaker/Muted"
		local callback = DEV_MSG.MUTE
		OneShotTimer.Add(150, "MILLISECONDS", function() urlGet(url, callback) end, "MUTE_TOGGLE") 
	 end        
end


function DEV_MSG.MUTE(value)
	Dbg:Trace("DEV_MSG.MUTE()")
	Dbg:Trace(value)

	local output = 0 	--mod 1000 value of Output Connection ID
	local state   		--Mute state represented as "True" or "False"
	
	--NOTE: values "01" & "00" will need to be modified based upon the deviced protocol specification

	if (value["muted"] == true) then 
		state = "True"
		--gMuteState[output] = "True"
	elseif (value["muted"] == false) then
		state = "False"
		--gMuteState[output] = "False"
	else
		Dbg:Info("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	if (gMuteState[output] ~= state) then
		NOTIFY.MUTE_CHANGED(output, state)
	end
	gMuteState[output] = state
end	

----------------------------------------- Protocol Helper Functions -----------------------------------------
function OnPowerOnDelayTimerExpired()
	Dbg:Trace("Power On Delay Timer expired")
	gPowerOnDelayTimer:KillTimer()
	gSendTimer:StartTimer(tonumber(Properties["Command Delay Milliseconds"]))
	if (gControlMethod == "IR") then 
		NOTIFY.ON()
		gPowerState = "ON"	
	end		
end

function GetCommand(idBinding, tParams, strCommandName)
	Dbg:Trace("GetCommand()")
	local t = {}
	
	if (gControlMethod == "IR") then
		t = CMDS_IR
	else
		t = CMDS
	end	

	if (t[strCommandName] ~= nil) then
		return t[strCommandName]
	else
		Dbg:Warn("GetCommand(): Command not defined - ".. strCommandName)
		return nil
	end	
end	

---------------------------------------------------------------------
-- Devices code
---------------------------------------------------------------------

function get_input_selection_commands()
	Dbg:Trace ("get_input_selection_commands()")
	local url = "/BeoZone/Zone/Sources"

	local callback = update_input_selection_commands
	urlGet(url, callback)

end

function update_input_selection_commands(info)
	Dbg:Trace ("update_input_selection_commands ()")
	local sourceAddress, source, source1
	for x,y in pairs(info.sources) do
		sourceAddress = y[1]
		source1 = sourceAddress
		source, pos = string.match (source1, "(.+):()")
		tInputCommandMap[source] = {activeSources = {primary = sourceAddress}}
	end	
	--PersistData["InputCommandMap"] = tInputCommandMap
end

function get_mac_address()
	local url = "/BeoDevice"
	local callback = update_mac_address
	urlGet(url, callback)	
end

function update_mac_address(info)
	local mac_address = info.beoDevice.hardware.mac
	C4:UpdateProperty("Device MAC Address", mac_address)
end

function format_mac_hex()
	local s = Properties["Device MAC Address"]
	local mac = s:gsub(":", " ")
	return tohex(mac)
end

---------------------------------------------------------------------
-- Power Functions
---------------------------------------------------------------------

function PRX_CMD.ON(idBinding, tParams)

	if (gPowerState == "ON") then 
		Dbg:Info("Power State is '" .. gPowerState .. "', exiting ON")
		return 
	end	
	
	if bUseWOL() then sendWOL() end
	
	if (tPowerCommandMap.on == nil) then 
		Dbg:Trace("Power On command not supported. Unit will power on with Source Selection.") 
	else
		local url = "/BeoDevice/powerManagement/standby"
		local tData = {standby = {powerState = "on"}}
		urlPut(url, tData, nil)
		
		url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
		local callback = DEV_MSG.VOLUME	
		urlGet(url, callback)	
		
		url = "/BeoDevice/powerManagement/standby"
		callback = DEV_MSG.POWER
		local command_delay = tonumber(Properties["Power On Delay Seconds"])
		OneShotTimer.Add(command_delay, "SECONDS", function() urlGet(url, callback) end, "PowerQuery") 
	end

end

function bUseWOL()
	local bUseWOL = false
	if (gIsTV) then
		--tv can use WOL
		if (Properties["Power On Method"] == "Wake On Lan") then bUseWOL = true end
	end	
	
	return bUseWOL
end

function sendWOL()
	C4:CreateNetworkConnection (6002, "255.255.255.255", 9) --create network connection "6002" for Wake on Lan
	C4:NetConnect(6002, 9, "UDP") --connect to connection 6002 and make it UDP for Wake On Lan

	local mac_address = format_mac_hex()
	local packet = string.rep(string.char(255), 6) .. string.rep(mac_address, 16) -- Build 'magic packet'. 
	--hexdump (packet) 
	C4:SendToNetwork (6002, 9, packet)	
		
	OneShotTimer.Add(300, "MILLISECONDS", function() C4:SendToNetwork (6002, 9, packet) end, "WoL")  --Added because we were sending it to fast, need delay in WoL
	OneShotTimer.Add(650, "MILLISECONDS", function() C4:SendToNetwork (6002, 9, packet) end, "WoL1")
	OneShotTimer.Add(1, "SECONDS", function() C4:SendToNetwork (6002, 9, packet) end, "WoL2")
	OneShotTimer.Add(5, "SECONDS", function() C4:SendToNetwork (6002, 9, packet) end, "WoL3")
end

function PRX_CMD.OFF(idBinding, tParams)
	if (idBinding ~= TV_PROXY_BINDINGID) then return end
	
	local url = "/BeoDevice/powerManagement/standby"
	local tData = {standby = {powerState = "standby"}}
	urlPut(url, tData, nil)
	
	--PowerQuery may not work well when using WOL. Timing is close between TVs responding with ON then OFF then not responding at all (OFFLINE). 
	gPowerState = "OFF"
	NOTIFY.OFF()	

	--local callback = DEV_MSG.POWER
	--local command_delay = 1 --tonumber(Properties["Power Off Delay Seconds"])
	--OneShotTimer.Add(command_delay, "SECONDS", function() urlGet(url, callback) end, "PowerQuery") 

end

function get_power_status()
	local url = "/BeoDevice/powerManagement/standby"
	local callback = DEV_MSG.POWER
	urlGet(url, callback)
end

function get_power_commands()
	Dbg:Trace ("get_power_commands()")
	local url = "/BeoDevice"
	local callback = get_power_managment_profile
	urlGet(url, callback)
end

function get_power_managment_profile(info)
	Dbg:Trace ("get_power_managment_profile()")
	local sourceAddress, source, source1
	for x,y in pairs(info.beoDevice.profiles) do
		if (y.name == "powerManagementProfile") then
			local s = string.match(y._links.self.href, "(%w+)")
			local url = "/BeoDevice/" .. s

			local callback = update_power_commands
			urlGet(url, callback)			
			
		end
	end	
end

function update_power_commands(info)
	Dbg:Trace ("update_power_commands()")
	for x,y in pairs(info.profile.powerManagement.standby._capabilities.value.powerState) do
		tPowerCommandMap[y] = y
	end
	--PersistData["PowerCommandMap"] = tPowerCommandMap
end


function get_stand_commands()
	Dbg:Trace ("get_stand_commands()")
	local url = "/BeoZone/Zone/Stand"

	local callback = update_stand_commands
	urlGet(url, callback)
end

function update_stand_commands(info)
	Dbg:Trace ("update_stand_commands()")
	tStandCommandMap = {}
	for x,y in pairs(info.stand.list) do
		tStandCommandMap[y.friendlyName] = {id = y.id, digit = y.digit}
	end
end

function get_speaker_group_commands()
	Dbg:Trace ("get_speaker_group_commands()")
	local url = "/BeoZone/Zone/Sound/SpeakerGroup"

	local callback = update_speaker_group_commands
	urlGet(url, callback)
end

function update_speaker_group_commands(info)
	Dbg:Trace ("update_speaker_group_commands()")
	tSpeakerGroupCommandMap = {}
	for x,y in pairs(info.speakerGroup.list) do
		tSpeakerGroupCommandMap[y.friendlyName] = {id = y.id, digit = y.digit}
	end
end

function get_picture_mode_commands()
	Dbg:Trace ("get_picture_mode_commands()")
	local url = "/BeoZone/Zone/Picture/Mode"

	local callback = update_picture_mode_commands
	urlGet(url, callback)
end

function update_picture_mode_commands(info)
	Dbg:Trace ("update_picture_mode_commands()")
	tPictureModeCommandMap = {}
	for x,y in pairs(info.mode.list) do
		tPictureModeCommandMap[y.friendlyName] = {id = y.id, digit = y.digit}
	end
end

function get_sound_mode_commands()
	Dbg:Trace ("get_sound_mode_commands()")
	local url = "/BeoZone/Zone/Sound/Mode"

	local callback = update_sound_mode_commands
	urlGet(url, callback)
end

function update_sound_mode_commands(info)
	Dbg:Trace ("update_sound_mode_commands()")
	tSoundModeCommandMap = {}
	for x,y in pairs(info.mode.list) do
		tSoundModeCommandMap[y.friendlyName] = {id = y.id, digit = y.digit}
	end
end

function get_picture_format_2d3d_commands()
	Dbg:Trace ("get_picture_format_2d3d_commands()")
	local url = "/BeoZone/Zone/Picture/Format/2d3d"

	local callback = update_picture_format_2d3d_commands
	urlGet(url, callback)
end

function update_picture_format_2d3d_commands(info)
	Dbg:Trace ("update_picture_format_2d3d_commands()")
	tPictureFormat2d3dCommandMap = {}
	for x,y in pairs(info["2d3d"].list) do
		tPictureFormat2d3dCommandMap[y.friendlyName] = {id = y.id}
	end
end

--called from 'Stand Position' Device Specific command to build list dialog
function GetStandPositionsForProgramming()
    Dbg:Trace('Getting Stand Positions...')
	
	local t = {}
	for j,k in pairs(tStandCommandMap) do
		t[j] = j
	end
    local list = getSortedList(t)

    if (table.maxn(list) == 0) then 
	   Dbg:Warn("Building Stand Positions List, close dialog window and then try again!")
	   table.insert (list, {text = "Getting list of Stand Positions from TV. Close dialog and then retry." , value = ""})
    end
	
	--refresh the list for next launch in the event that the data has changed on the TV
	get_stand_commands()

    return list
	
end

--called from 'Speaker Group' Device Specific command to build list dialog
function GetSpeakerGroupsForProgramming()
    Dbg:Trace('Getting Speaker Groups...')
	
	local t = {}
	for j,k in pairs(tSpeakerGroupCommandMap) do
		t[j] = j
	end
    local list = getSortedList(t)

    if (table.maxn(list) == 0) then 
	   Dbg:Warn("Building Stand Positions List, close dialog window and then try again!")
	   table.insert (list, {text = "Getting list of Speaker Groups from TV. Close dialog and then retry." , value = ""})
    end

	--refresh the list for next launch in the event that the data has changed on the TV
	get_speaker_group_commands()
	
    return list
	
end

--called from 'Picture Mode' Device Specific command to build list dialog
function GetPictureModesForProgramming()
    Dbg:Trace('Getting Picture Modes...')
    
	local t = {}
	for j,k in pairs(tPictureModeCommandMap) do
		t[j] = j
	end
    local list = getSortedList(t)

    if (table.maxn(list) == 0) then 
	   Dbg:Warn("Building Picture Modes List, close dialog window and then try again!")
	   table.insert (list, {text = "Getting list of Picture Modes from TV. Close dialog and then retry." , value = ""})
    end

	--refresh the list for next launch in the event that the data has changed on the TV
	get_picture_mode_commands()
	
    return list
	
end

--called from 'Sound Mode' Device Specific command to build list dialog
function GetSoundModesForProgramming()
    Dbg:Trace('Getting Sound Modes...')
    
	local t = {}
	for j,k in pairs(tSoundModeCommandMap) do
		t[j] = j
	end
    local list = getSortedList(t)

    if (table.maxn(list) == 0) then 
	   Dbg:Warn("Building Sound Modes List, close dialog window and then try again!")
	   table.insert (list, {text = "Getting list of Sound Modes from TV. Close dialog and then retry." , value = ""})
    end

	--refresh the list for next launch in the event that the data has changed on the TV
	get_sound_mode_commands()
	
    return list

end

--called from 'Picture Format 2D3D' Device Specific command to build list dialog
function GetPictureFormat2d3dForProgramming()
    Dbg:Trace('Getting Picture Format 2D3D...')
    
	local t = {}
	for j,k in pairs(tPictureFormat2d3dCommandMap) do
		t[j] = j
	end
    local list = getSortedList(t)

    if (table.maxn(list) == 0) then 
	   Dbg:Warn("Building Picture Format 2D3D List, close dialog window and then try again!")
	   table.insert (list, {text = "Getting list of Picture Format 2D3D from TV. Close dialog and then retry." , value = ""})
    end

	--refresh the list for next launch in the event that the data has changed on the TV
	get_picture_format_2d3d_commands()
	
    return list
	
end

function getSortedList(t)
	local a = {}
	for j in pairs(t) do 
		table.insert(a,j)
	end 
	table.sort(a)
    
    local list = {}

    for i = 1,table.maxn(a) do
        local id = a[i]
		table.insert (list, {text = id, value = t[id]})
    end
	
    return list
end


---------------------------------------------------------------------
-- TV Proxy Tuner Functions
---------------------------------------------------------------------
function PRX_CMD.PULSE_CH_UP(idBinding, tParams)
	local url = "/BeoZone/Zone/List/StepUp"	
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_CH_UP(idBinding, tParams)
	local url = "/BeoZone/Zone/List/StepUp"
	local tData = {ToBeReleased = true}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_CH_UP(idBinding, tParams)
	local cmd = "/BeoZone/Zone/List/StepUp"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.PULSE_CH_DOWN(idBinding, tParams)
	local url = "/BeoZone/Zone/List/StepDown"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_CH_DOWN(idBinding, tParams)
	local url = "/BeoZone/Zone/List/StepDown"
	local tData = {ToBeReleased = true}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_CH_DOWN(idBinding, tParams)
	local cmd = "/BeoZone/Zone/List/StepDown"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.SET_CHANNEL(idBinding, tParams)
	local channel= tParams['CHANNEL']
	Dbg:Trace ("PRX_CMD.SET_CHANNEL(channel = " .. channel .. ")")
	
	local url = "/BeoZone/Zone/Digits "
	local tData = {}	
	for i=1,string.len(channel) do
		local c = string.sub(channel,i,i)
		if (string.find(c, "%D")) then
			Dbg:Warn("PRX_CMD.SET_CHANNEL: delimiter '" .. c .. "' not supported!")
			-- TODO: change delimiter command below if it is not a DOT...
			--SEND_COMMAND_FROM_COMMAND_TABLE(idBinding, "DOT")
		else
			tData = {digits = tonumber(c),timestamp = 0}	
			OneShotTimer.Add(150, "MILLISECONDS", function() urlPost(url, tData, nil) end, "Set Channel, digit " .. c)
		end
	end 	
	NOTIFY.CHANNEL_CHANGED(channel)
end
---------------------------------------------------------------------
-- Input Selection Functions
---------------------------------------------------------------------
gLastSelectedInput = ""
function PRX_CMD.SET_INPUT(idBinding, tParams)
	gSetInputInProgressTimer:StartTimer(30, "SECONDS")
	
	local input = tonumber(tParams["INPUT"] % 1000)
	gLastSelectedInput = input
  	local callback_input = DEV_MSG.INPUT
	local url = "/BeoZone/Zone/ActiveSourceType"	
	local tData = tInputCommandMap[tInputConnMapByID[input].Name]
	
	urlPost(url, tData, nil) 
	OneShotTimer.Add(2, "SECONDS", function() urlGet(url, callback_input) end, "GetInput ActiveSources") 
	
	if (gPowerState ~= "ON") then 
		OneShotTimer.Add(3, "SECONDS", function() getVolumeLevel() end, "getVolumeLevel") 
		OneShotTimer.Add(4, "SECONDS", function() get_power_status() end, "get_power_status") 
	end	
end
 
function OnRecoverInputSelectionTimerExpired()
	Dbg:Trace("OnRecoverInputSelectionTimerExpired()")
	gRecoverInputSelectionTimer:KillTimer()
	if (gLastSelectedInput == "" or gLastSelectedInput == nil) then return end

  	local callback_input = DEV_MSG.INPUT
	local url = "/BeoZone/Zone/ActiveSourceType"	
	local tData = tInputCommandMap[tInputConnMapByID[gLastSelectedInput].Name]
	
    urlPost(url, tData, nil) 
    OneShotTimer.Add(1, "SECONDS", function() urlGet(url, callback_input) end, "GetInput ActiveSources") 
	OneShotTimer.Add(2, "SECONDS", function() getVolumeLevel() end, "getVolumeLevel") 
	OneShotTimer.Add(3, "SECONDS", function() get_power_status() end, "get_power_status") 
end

function OnSetInputInProgressTimerExpired()
	Dbg:Trace("OnSetInputInProgressTimerExpired()")
	gSetInputInProgressTimer:KillTimer()
	
end

---------------------------------------------------------------------
-- Menu Functions
---------------------------------------------------------------------

function PRX_CMD.INPUT_TOGGLE(idBinding, tParams)
	QueueCommand(CMDS["INPUT_TOGGLE"])
end

function PRX_CMD.INFO(idBinding, tParams)
	local url = "/BeoZone/Zone/Device/Information"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.GUIDE(idBinding, tParams)
	local url = "/BeoZone/Zone/Menu/ElectronicProgramGuide"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Stop"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.SKIP_REV(idBinding, tParams)
	local url = ""
	if (idBinding == RADIO_PROXY_BINDINGID or idBinding == DVBRADIO_PROXY_BINDINGID or idBinding == SAT_PROXY_BINDINGID ) then
		url = "/BeoZone/Zone/List/StepDown"
	else
		url = "/BeoZone/Zone/Stream/Backward"
	end
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.SKIP_FWD(idBinding, tParams)
	local url = ""
	if (idBinding == RADIO_PROXY_BINDINGID or idBinding == DVBRADIO_PROXY_BINDINGID or idBinding == SAT_PROXY_BINDINGID ) then
		url = "/BeoZone/Zone/List/StepUp"
	else
		url = "/BeoZone/Zone/Stream/Forward"
	end
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.SCAN_FWD(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Wind"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_SCAN_FWD(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Wind"
	local tData = {ToBeReleased = true}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_SCAN_FWD(idBinding, tParams)
	local cmd = "/BeoZone/Zone/Stream/Wind"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.SCAN_REV(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Rewind"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_SCAN_REV(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Rewind"
	local tData = {ToBeReleased = true}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_SCAN_REV(idBinding, tParams)
	local cmd = "/BeoZone/Zone/Stream/Rewind"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.PLAY(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Play"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.PAUSE(idBinding, tParams)
	local url = "/BeoZone/Zone/Stream/Pause"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.RECORD(idBinding, tParams)
	local url = "/BeoZone/Zone/Record/Record"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.MENU(idBinding, tParams)
	local url = "/BeoZone/Zone/Menu/Root"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.CANCEL(idBinding, tParams) --This might need to be the EXIT command
	local url = "/BeoZone/Zone/Cursor/Back"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.PAGE_UP(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/PageUp"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.PAGE_DOWN(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/PageDown"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.UP(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Up"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_UP(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Up"
	local tData = {ToBeReleased = true}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_UP(idBinding, tParams)
	local cmd = "/BeoZone/Zone/Cursor/Up"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.DOWN(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Down"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_DOWN(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Down"
	local tData = {ToBeReleased = true}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_DOWN(idBinding, tParams)
	local cmd = "/BeoZone/Zone/Cursor/Down"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.LEFT(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Left"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_LEFT(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Left"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_LEFT(idBinding, tParams)
	local cmd = "/BeoZone/Zone/Cursor/Left"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.RIGHT(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Right"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.START_RIGHT(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Right"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.STOP_RIGHT(idBinding, tParams)
	local cmd = "/BeoZone/Zone/Cursor/Right"
	local delay = 10
	RELEASE(cmd, delay)
end

function PRX_CMD.ENTER(idBinding, tParams)
	local url = "/BeoZone/Zone/Cursor/Select"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.RECALL(idBinding, tParams)
	local url = "/BeoZone/Zone/List/Repeat"	--PreviousElement
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.OPEN_CLOSE(idBinding, tParams)
end

function PRX_CMD.PROGRAM_A(idBinding, tParams)
	local url = "/BeoZone/Zone/Generic/Red"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.PROGRAM_B(idBinding, tParams)
	local url = "/BeoZone/Zone/Generic/Green"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.PROGRAM_C(idBinding, tParams)
	local url = "/BeoZone/Zone/Generic/Yellow"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end

function PRX_CMD.PROGRAM_D(idBinding, tParams)
	local url = "/BeoZone/Zone/Generic/Blue"
	local tData = {ToBeReleased = false}
	urlPost(url, tData, nil)
end
---------------------------------------------------------------------
-- Digit Functions
---------------------------------------------------------------------
function PRX_CMD.NUMBER_0(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 0,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_1(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 1,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_2(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 2,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_3(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 3,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_4(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 4,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_5(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 5,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_6(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 6,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_7(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 7,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_8(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 8,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.NUMBER_9(idBinding, tParams)
	local url = "/BeoZone/Zone/Digits"
	local tData = {digits = 9,timestamp = 0}
	urlPost(url, tData, nil)
end

function PRX_CMD.STAR(idBinding, tParams)
	--QueueCommand(CMDS["STAR"])
	QueueCommand(GetCommand(idBinding, tParams, "STAR"))
end

function PRX_CMD.POUND(idBinding, tParams)
	--QueueCommand(CMDS["POUND"])
	QueueCommand(GetCommand(idBinding, tParams, "POUND"))
end

---------------------------------------------------------------------
-- Volume Control Code
---------------------------------------------------------------------
function PRX_CMD.MUTE_OFF(idBinding, tParams)
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/Muted"
	local tData = {muted = false}
	urlPut(url, tData, nil)
	
	local callback = DEV_MSG.MUTE
	OneShotTimer.Add(150, "MILLISECONDS", function() urlGet(url, callback) end, "MUTE_TOGGLE")  --Added because we were sending it to fast

end

function PRX_CMD.MUTE_ON(idBinding, tParams)
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/Muted"
	local tData = {muted = true}
	urlPut(url, tData, nil)
	
	local callback = DEV_MSG.MUTE
	OneShotTimer.Add(150, "MILLISECONDS", function() urlGet(url, callback) end, "MUTE_TOGGLE")  --Added because we were sending it to fast
end

function PRX_CMD.MUTE_TOGGLE(idBinding, tParams)
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/Muted"
	local tData
	local output = 0	--tParams.OUTPUT % 1000
	
	if (gMuteState[output] == "True") then
		tData = {muted = false}
	else
		tData = {muted = true}
	end
	urlPut(url, tData, nil)
	
	local callback = DEV_MSG.MUTE
	OneShotTimer.Add(150, "MILLISECONDS", function() urlGet(url, callback) end, "MUTE_TOGGLE")  --Added because we were sending it to fast
end

function PRX_CMD.GET_VOLUME_LEVEL()
	--If no Power On commands then WOL will be used. Volume status will be set by OnRecoverInputSelectionTimerExpired().
	if (tPowerCommandMap.on == nil) then return end
	getVolumeLevel()
end

function getVolumeLevel()
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"
	local callback = DEV_MSG.VOLUME
	urlGet(url, callback)
end

function PRX_CMD.SET_VOLUME_LEVEL(idBinding, tParams)
	local output = 0
	local volumeLevel = tonumber(tParams['LEVEL'])

	--todo  Test volume scale and 0-90
	if volumeLevel > 90 then 
		volumeLevel = 90
	end
	
	--Edit "minDeviceLevel" & "maxDeviceLevel" values, based upon the protocol specification for volume range
	local minDeviceLevel = 0
	local maxDeviceLevel = 100
	volumeLevel = ConvertVolumeToDevice(volumeLevel, minDeviceLevel, maxDeviceLevel)	
	
	Dbg:Info('volumeLevel: ' .. volumeLevel)

	local url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"
	local tData = {level = (volumeLevel)}
	local callback = DEV_MSG.VOLUME
	url2 = "/BeoZone/Zone/Sound/Volume/Speaker/Muted"
	local callback2 = DEV_MSG.MUTE
	
	if (gPowerState == "OFF") then 
		-- This is the power on delay
		Dbg:Trace "SET_VOLUME_LEVEL was called when the gPowerState = OFF. Delay using the Property 'Power On Delay Seconds'"
		
		OneShotTimer.Add(tonumber(Properties["Power On Delay Seconds"]), "SECONDS",function() urlPut(url, tData, nil) end, "SetVolume") 
		
		OneShotTimer.Add(tonumber(Properties["Power On Delay Seconds"])+3, "SECONDS", function() urlGet(url, callback) end, "VolCallback")  

		OneShotTimer.Add(tonumber(Properties["Power On Delay Seconds"])+4, "SECONDS", function() urlGet(url2, callback2) end, "MUTE_TOGGLE")
		return 
	end		

	OneShotTimer.Add(250, "MILLISECONDS", function() urlPut(url, tData, nil) end, "SetVolume")  --Added because we were sending it to fast
		
	OneShotTimer.Add(500, "MILLISECONDS", function() urlGet(url, callback) end, "VolCallback")  --Added because we were sending it to fast
	
	OneShotTimer.Add(650, "MILLISECONDS", function() urlGet(url2, callback2) end, "MUTE_TOGGLE")
	
end

function PRX_CMD.PULSE_VOL_DOWN(idBinding, tParams)
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/ContinuousLevelAction"
	local tData = {continuousLevelAction = "continuousDown", continuousLevelTimeoutAction = 0}	--continuousLevelTimeoutAction = 100
	urlPost(url, tData, nil)
	
	url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
	local callback = DEV_MSG.VOLUME	
	OneShotTimer.Add(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS",function() urlGet(url, callback) end, "VolCallback")
end

function PRX_CMD.PULSE_VOL_UP(idBinding, tParams)
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/ContinuousLevelAction"
	local tData = {continuousLevelAction = "continuousUp", continuousLevelTimeoutAction = 0}	--continuousLevelTimeoutAction = 100
	urlPost(url, tData, nil)
	
	url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
	local callback = DEV_MSG.VOLUME	
	OneShotTimer.Add(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", function() urlGet(url, callback) end, "VolCallback")
end

function PRX_CMD.START_VOL_DOWN(idBinding, tParams)
	ChangeVolume("START_VOL_DOWN")
--[[
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/ContinuousLevelAction"
	local tData = {continuousLevelAction = "continuousDown"}
	urlPost(url, tData, nil)
	
	url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
	local callback = DEV_MSG.VOLUME	
	OneShotTimer.Add(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS",function() urlGet(url, callback) end, "VolCallback")
--]]
end

function PRX_CMD.START_VOL_UP(idBinding, tParams)
	ChangeVolume("START_VOL_UP")
--[[
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/ContinuousLevelAction"
	local tData = {continuousLevelAction = "continuousUp"}
	urlPost(url, tData, nil)
	
	url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
	local callback = DEV_MSG.VOLUME	
	OneShotTimer.Add(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", function() urlGet(url, callback) end, "VolCallback")
--]]
end

function PRX_CMD.STOP_VOL_DOWN(idBinding, tParams)
	ChangeVolume("STOP_VOL_DOWN")
--[[
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/ContinuousLevelAction"
	local tData = {continuousLevelAction = "none"}
	urlPost(url, tData, nil)
	
	url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
	local callback = DEV_MSG.VOLUME	
	OneShotTimer.Add(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", function() urlGet(url, callback) end, "VolCallback")
--]]
end

function PRX_CMD.STOP_VOL_UP(idBinding, tParams)
	ChangeVolume("STOP_VOL_UP")
--[[
	local url = "/BeoZone/Zone/Sound/Volume/Speaker/ContinuousLevelAction"
	local tData = {continuousLevelAction = "none"}
	urlPost(url, tData, nil)
	
	url = "/BeoZone/Zone/Sound/Volume/Speaker/Level"	
	local callback = DEV_MSG.VOLUME	
	OneShotTimer.Add(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS", function() urlGet(url, callback) end, "VolCallback")
--]]
end

-------------------------------------- Volume Helper Functions --------------------------------------
function ChangeVolume(strCommand)
	local output = 0
	if (strCommand == 'STOP_VOL_UP') or (strCommand == 'STOP_VOL_DOWN') then
		gVolumeIsRamping = false
		gVolumeRamping[output].state = false
		gVolumeRamping[output].mode = ""
	elseif (strCommand == 'START_VOL_UP') then 
		gVolumeIsRamping = true
		gVolumeRamping[output].state = true
		gVolumeRamping[output].mode = "VOLUME_UP" 
		PRX_CMD.PULSE_VOL_UP(5001, {})
	elseif (strCommand == 'START_VOL_DOWN') then 	
		gVolumeIsRamping = true
		gVolumeRamping[output].state = true
		gVolumeRamping[output].mode = "VOLUME_DOWN"
		PRX_CMD.PULSE_VOL_DOWN(5001, {})
	else
		Dbg:Warn(strCommand .. " not handled in ChangeVolume()")
	end
end

function ContinueVolumeRamping()
	local output = 0
    if (gVolumeRamping[output].mode == "VOLUME_DOWN") then
		PRX_CMD.PULSE_VOL_DOWN(5001,{})
	elseif (gVolumeRamping[output].mode == "VOLUME_UP") then	
		PRX_CMD.PULSE_VOL_UP(5001,{})                                
	else
		Dbg:Warn("ContinueVolumeRamping(), Invalid Mode")
		return
	end

end

function GetNextVolumeCurveValue(output, volume)
	local i, point
	volume=tonumber(volume)
	if (gVolumeRamping[output].mode == "VOLUME_UP") then
		for i=1,table.maxn(tVolumeCurve) do
			point=tonumber(tVolumeCurve[i])
			if point > volume then		
				return tVolumeCurve[i]
			end
		end
	elseif (gVolumeRamping[output].mode == "VOLUME_DOWN") then
		for i=table.maxn(tVolumeCurve),1,-1 do
			point=tonumber(tVolumeCurve[i])
			if point < volume then
				return tVolumeCurve[i]
			end
		end
	else
		Dbg:Warn("Volume Ramping Mode not set for "  .. tOutputConnMap[output])
		return nil
	end 
end

function ConvertVolumeToC4(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a device to a 
	--percentage value that can be used by C4 proxies
	--"volume" is the volume value from the device
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, minDeviceLevel, maxDeviceLevel, 0, 100)
end

function ConvertVolumeToDevice(volume, minDeviceLevel, maxDeviceLevel)
	--to be used when converting a volume level from a C4 proxy to a 
	--value that can be used by the device 
	--"volume" is the volume value from the C4 proxy
	--"minDeviceLevel" & "maxDeviceLevel" are the minimum and maximum volume levels
	--as specified in the device protocol documentation
	return ProcessVolumeLevel(volume, 0, 100, minDeviceLevel, maxDeviceLevel)
end

function ProcessVolumeLevel(volLevel, minVolLevel, maxVolLevel, minDeviceLevel, maxDeviceLevel)
	local level = (volLevel-minVolLevel)/(maxVolLevel-minVolLevel)
	  --Dbg:Info("level = " .. level)
	  local vl=(level*(maxDeviceLevel-minDeviceLevel))+minDeviceLevel
	  --Dbg:Info("vl = " .. vl)
	  vl= tonumber(("%.".."0".."f"):format(vl))
	  --Dbg:Info("vl new = " .. vl)
	  Dbg:Info("ProcessVolumeLevel(level in=" .. volLevel .. ", level out=" .. vl .. ")")
	  return vl
end


------------------------------ Power Notification Functions ------------------------------
function NOTIFY.ON()
	SendSimpleNotify("ON")
	Dbg:Trace ("NOTIFY.ON ()")
end

function NOTIFY.OFF()
	SendSimpleNotify("OFF")
	Dbg:Trace ("NOTIFY.OFF ()")
end

------------------------------ Input Selection Notification Code ------------------------------
function NOTIFY.INPUT_CHANGED(input)
	Dbg:Trace ("NOTIFY.INPUT_CHANGED()" ..input)
	C4:SendToProxy(TV_PROXY_BINDINGID, "INPUT_CHANGED", {INPUT = input})
end
------------------------------ Channel Notification Code ------------------------------
function NOTIFY.CHANNEL_CHANGED(channel)
	C4:SendToProxy(TV_PROXY_BINDINGID, "CHANNEL_CHANGED", {CHANNEL = channel})
end
function NOTIFY.MUTE_CHANGED(output, state)
	Dbg:Trace ("NOTIFY.MUTE_CHANGED state = "..state)
	C4:SendToProxy(TV_PROXY_BINDINGID, "MUTE_CHANGED", {MUTE = state, OUTPUT = output + 4000})
end

function NOTIFY.VOLUME_LEVEL_CHANGED(output, level)
	Dbg:Trace ("NOTIFY.VOLUME_LEVEL_CHANGED level =   "..level)
	SendToProxy(DEFAULT_PROXY_BINDINGID, "VOLUME_LEVEL_CHANGED", {LEVEL = level, OUTPUT = output + 4000})
end


-- NOTE: OneShotTimer Must be at end of code.
---------------------------------------------------------
--------------  OneShotTimer Timers Module  ----------------
---------------------------------------------------------
do
  local OTE, ODD
  local MTFUNC = function() end

  OneShotTimer = {}
  local tCallback = {}
  
  function OneShotTimer.OnTimerExpired(idTimer)
    if (tCallback[idTimer] ~= nil) then
      C4:KillTimer(idTimer)
      tCallback[idTimer].CALLBACK(tostring(tCallback[idTimer].NAME))
      tCallback[idTimer] = nil
    else
      OTE(idTimer)
    end
  end

  function OneShotTimer.ClearAll()
    for k,v in pairs(tCallback) do C4:KillTimer(k) end
    tCallback = {}
  end

  function OneShotTimer.OnDriverDestroyed()
    OneShotTimer.ClearAll()
    ODD()
  end

  function OneShotTimer.Kill(Name)
    if (Name ~= nil) then
      for k,v in pairs(tCallback) do if (v.NAME == Name) then C4:KillTimer(k) tCallback[k] = nil end end
    end
  end

  function OneShotTimer.Add(nInterval, strUnits, fCallback, Name)
    local id = C4:AddTimer(nInterval, strUnits)

    -- Look for name if not nil, if found, remove existing timer callback...
    if (Name ~= nil) then
      for k,v in pairs(tCallback) do if (v.NAME == Name) then C4:KillTimer(k) tCallback[k] = nil end end
    end
    tCallback[id] = {CALLBACK = fCallback, NAME = Name}
  end

  -----  Overrides:  --------
  OTE , OnTimerExpired = OnTimerExpired or MTFUNC, OneShotTimer.OnTimerExpired
  ODD , OnDriverDestroyed = OnDriverDestroyed or MTFUNC, OneShotTimer.OnDriverDestroyed
  ---  END Overrides:  ------
end
---------------------------------------------------------
-----------  END OneShotTimer Timers Module  ---------------
---------------------------------------------------------